#  Copyright 2020 United Kingdom Research and Innovation
#  Copyright 2020 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
import sys
import unittest
from unittest.mock import patch
from utils import initialise_tests

import numpy as np
import os
from cil.framework import ImageGeometry
from cil.framework.labels import AngleUnit
from cil.io import NEXUSDataReader, NikonDataReader, ZEISSDataReader
from cil.io import TIFFWriter, TIFFStackReader
from cil.io.utilities import HDF5_utilities
from cil.processors import Slicer
from utils import has_astra, has_nvidia
from cil.utilities.quality_measures import mse
from cil.utilities import dataexample
import logging
import glob
import json
from cil.io import utilities
from cil.io import RAWFileWriter
import configparser
import tempfile


initialise_tests()

has_dxchange = True
try:
    import dxchange
except ImportError as ie:
    has_dxchange = False
has_olefile = True
try:
    import olefile
except ImportError as ie:
    has_olefile = False
has_wget = True
try:
    import wget
except ImportError as ie:
    has_wget = False
if has_astra:
    from cil.plugins.astra import FBP


has_file = False
has_recon_file = False
basedir = dataexample.REMOTEDATA.CIL_DATA_DIR
if basedir is not None:
    dataexample.WALNUT.download_data(data_dir=basedir, prompt=False)

    test_txrm_file = os.path.join(basedir, "walnut/valnut/valnut_2014-03-21_643_28/tomo-A/", "valnut_tomo-A.txrm")
    # strip double quotes if they exist
    test_txrm_file = test_txrm_file.strip('"')
    test_txrm_file = os.path.abspath(test_txrm_file)
    has_file = os.path.isfile(test_txrm_file)
    
    test_3d_recon_file = os.path.join(basedir, "walnut/valnut/valnut_2014-03-21_643_28/tomo-A/", "valnut_tomo-A_recon.txm")
    # strip double quotes if they exist
    test_3d_recon_file = test_3d_recon_file.strip('"')
    test_3d_recon_file = os.path.abspath(test_3d_recon_file)
    has_recon_file = os.path.isfile(test_3d_recon_file)


# Change the level of the logger to WARNING (or whichever you want) to see more information
logging.basicConfig(level=logging.WARNING)
log = logging.getLogger(__name__)
log.info("has_astra %s", has_astra)
log.info("has_wget %s", has_wget)
log.info("has_olefile %s", has_olefile)
log.info("has_dxchange %s", has_dxchange)
log.info("has_file %s", has_file)
if not has_file:
    log.info("This unittest requires the walnut Zeiss dataset saved in %s", basedir)

  

class TestZeissDataReader(unittest.TestCase):
    
    @unittest.skipIf(not (has_file and has_olefile and has_dxchange), 
                     f"Missing prerequisites: has_file {has_file}, has_olefile {has_olefile} has_dxchange {has_dxchange}")
    def test_geometry_gpu(self):
        
        reader = ZEISSDataReader()           
        reader.set_up(file_name=test_txrm_file)

        geometry = reader.get_geometry()
        # print (geometry)
        assert geometry is not None
        assert geometry.geom_type == 'CONE'
        assert geometry.dimension == '3D'
        
        from cil.framework import AcquisitionGeometry
        metadata = reader.get_metadata()
        _geometry = AcquisitionGeometry.create_Cone3D(
                [0,-metadata['dist_source_center'],0],[0,metadata['dist_center_detector'],0] \
                ) \
                    .set_panel([metadata['image_width'], metadata['image_height']],\
                        pixel_size=[metadata['detector_pixel_size']/1000,metadata['detector_pixel_size']/1000])\
                    .set_angles(metadata['thetas'],angle_unit=AngleUnit.RADIAN)
        
        assert _geometry == geometry


    @unittest.skipIf(not (has_file and has_olefile and has_dxchange and has_recon_file), 
                     f"Missing prerequisites: has_file {has_file}, has_recon_file {has_recon_file} has_olefile {has_olefile} has_dxchange {has_dxchange}, has_astra {has_astra} has_wget {has_wget}")
    def test_read_txm_recon_file_gpu(self):
        zreader = ZEISSDataReader()           
        zreader.set_up(file_name=test_3d_recon_file)

        metadata = zreader.get_metadata()
        
        _geometry = ImageGeometry(voxel_num_x=metadata['image_width'], 
                                  voxel_num_y=metadata['image_height'],
                                  voxel_num_z=metadata['number_of_images'],
                                  voxel_size_x=metadata['pixel_size'],
                                  voxel_size_y=metadata['pixel_size'],
                                  voxel_size_z=metadata['pixel_size'],
                                 )
        data3d = zreader.read()

        assert _geometry == data3d.geometry


    @unittest.skipIf(not (has_file and has_olefile and has_dxchange ), 
                     f"Missing prerequisites: has_file {has_file}, has_olefile {has_olefile} has_dxchange {has_dxchange}, has_astra {has_astra} has_wget {has_wget}")
    def test_read_and_reconstruct_2D_gpu(self):

        # Here we may read the TXM file that comes with the Walnut Zeiss dataset, extract
        # the central slice and compare it with the FBP reconstruction from the data.
        # However, https://github.com/TomographicImaging/CIL/issues/2175        
        # So at the moment we just read the file and reconstruct it.
        
        zreader = ZEISSDataReader()           
        zreader.set_up(file_name=test_txrm_file)
        data = zreader.read()

        # get central slice
        data2d = data.get_slice(vertical='centre')
        # neg log
        data2d.log(out=data2d)
        data2d *= -1

        ig2d = data2d.geometry.get_ImageGeometry()
        assert data2d.geometry is not None, "data2d geometry is None"
        fbpalg = FBP(ig2d,data2d.geometry)
        fbpalg.set_input(data2d)

        recfbp = fbpalg.get_output()

    def test_file_not_found_error(self):
        reader = ZEISSDataReader()           
        
        with self.assertRaises(FileNotFoundError):
            reader.set_up(file_name='no-file')

    @unittest.skipIf(has_dxchange, f"This unit test runs only if dxchange is not installed: has_dxchange {has_dxchange}")
    def test_import_error(self):
        reader = ZEISSDataReader()
        with self.assertRaises(ImportError):
            reader.set_up(file_name=test_txrm_file)

class TestTIFF(unittest.TestCase):
    def setUp(self) -> None:
        self.TMP = tempfile.TemporaryDirectory()
        self.cwd = os.path.join(self.TMP.name, 'rawtest')

    def tearDown(self) -> None:
        self.TMP.cleanup()

    def get_slice_imagedata(self, data):
        '''Returns only 2 slices of data'''
        # data = dataexample.SIMULATED_SPHERE_VOLUME.get()
        data.dimension_labels[0]
        roi = {data.dimension_labels[0]: (0,2,1),
               data.dimension_labels[1]: (None, None, None),
               data.dimension_labels[2]: (None, None, None)}
        return Slicer(roi=roi)(data)


    def test_tiff_stack_ImageData(self):
        # data = dataexample.SIMULATED_SPHERE_VOLUME.get()
        data = self.get_slice_imagedata(
            dataexample.SIMULATED_SPHERE_VOLUME.get()
        )

        fname = os.path.join(self.cwd, "unittest")

        writer = TIFFWriter(data=data, file_name=fname)
        writer.write()

        reader = TIFFStackReader(file_name=self.cwd)
        read_array = reader.read()

        np.testing.assert_allclose(data.as_array(), read_array)

        read = reader.read_as_ImageData(data.geometry)
        np.testing.assert_allclose(data.as_array(), read.as_array())


    def test_tiff_stack_AcquisitionData(self):
        # data = dataexample.SIMULATED_CONE_BEAM_DATA.get()
        data = self.get_slice_imagedata(
            dataexample.SIMULATED_CONE_BEAM_DATA.get()
        )


        fname = os.path.join(self.cwd, "unittest")

        writer = TIFFWriter(data=data, file_name=fname)
        writer.write()

        reader = TIFFStackReader(file_name=self.cwd)
        read_array = reader.read()

        np.testing.assert_allclose(data.as_array(), read_array)

        read = reader.read_as_AcquisitionData(data.geometry)
        np.testing.assert_allclose(data.as_array(), read.as_array())


    def test_tiff_stack_ImageDataSlice(self):
        data = dataexample.SIMULATED_SPHERE_VOLUME.get()

        fname = os.path.join(self.cwd, "unittest")

        writer = TIFFWriter(data=data, file_name=fname)
        writer.write()

        roi = {'axis_0': -1, 'axis_1': -1, 'axis_2': (None, None, 2)}

        reader = TIFFStackReader(file_name=self.cwd, roi=roi, mode='slice')
        read_array = reader.read()

        shape = [el for el in data.shape]
        shape[2] /= 2

        np.testing.assert_allclose(shape, read_array.shape )

        roi = {'axis_0': (0, 2, None), 'axis_1': -1, 'axis_2': -1}

        reader = TIFFStackReader(file_name=self.cwd, roi=roi, mode='slice')
        read_array = reader.read()

        np.testing.assert_allclose(data.as_array()[:2], read_array)


    def test_tiff_stack_ImageData_wrong_file(self):
        # data = dataexample.SIMULATED_SPHERE_VOLUME.get()
        data = self.get_slice_imagedata(
            dataexample.SIMULATED_SPHERE_VOLUME.get()
        )

        fname = os.path.join(self.cwd, "unittest")

        writer = TIFFWriter(data=data, file_name=fname)
        writer.write()

        for el in glob.glob(os.path.join(self.cwd , "unittest*.tiff")):
            # print (f"modifying {el}")
            with open(el, 'w') as f:
                f.write('BOOM')
            break

        reader = TIFFStackReader(file_name=self.cwd)
        try:
            read_array = reader.read()
            assert False
        except:
            assert True

    def test_TIFF_compression3D_0(self):
        self.TIFF_compression_test(None)

    def test_TIFF_compression3D_1(self):
        self.TIFF_compression_test('uint8')

    def test_TIFF_compression3D_2(self):
        self.TIFF_compression_test('uint16')

    def test_TIFF_compression3D_3(self):
        with self.assertRaises(ValueError) as context:
            self.TIFF_compression_test('whatever_compression')

    def test_TIFF_compression4D_0(self):
        self.TIFF_compression_test(None,2)

    def test_TIFF_compression4D_1(self):
        self.TIFF_compression_test('uint8',2)

    def test_TIFF_compression4D_2(self):
        self.TIFF_compression_test('uint16',2)

    def test_TIFF_compression4D_3(self):
        with self.assertRaises(ValueError) as context:
            self.TIFF_compression_test('whatever_compression',2)

    def TIFF_compression_test(self, compression, channels=1):
        X=4
        Y=5
        Z=6
        C=channels
        if C == 1:
            ig = ImageGeometry(voxel_num_x=4, voxel_num_y=5, voxel_num_z=6)
        else:
            ig = ImageGeometry(voxel_num_x=4, voxel_num_y=5, voxel_num_z=6, channels=C)
        data = ig.allocate(0)
        data.fill(np.arange(X*Y*Z*C).reshape(ig.shape))

        compress = utilities.get_compress(compression)
        dtype = utilities.get_compressed_dtype(data.array, compression)
        scale, offset = utilities.get_compression_scale_offset(data.array, compression)
        if C > 1:
            assert data.ndim == 4
        fname = os.path.join(self.cwd, "unittest")
        writer = TIFFWriter(data=data, file_name=fname, compression=compression)
        writer.write()
        # force the reader to use the native TIFF dtype by setting dtype=None
        reader = TIFFStackReader(file_name=self.cwd, dtype=None)
        read_array = reader.read()
        if C > 1:
            read_array = reader.read_as_ImageData(ig).array


        if compress:
            tmp = data.array * scale + offset
            tmp = np.asarray(tmp, dtype=dtype)
            # test if the scale and offset are written to the json file
            with open(os.path.join(self.cwd, "scaleoffset.json"), 'r') as f:
                d = json.load(f)
            assert d['scale'] == scale
            assert d['offset'] == offset
            # test if the scale and offset are read from the json file
            sc, of = reader.read_scale_offset()
            assert sc == scale
            assert of == offset

            recovered_data = (read_array - of)/sc
            np.testing.assert_allclose(recovered_data, data.array, rtol=1e-1, atol=1e-2)

            # test read_rescaled
            approx = reader.read_rescaled()
            np.testing.assert_allclose(approx.ravel(), data.array.ravel(), rtol=1e-1, atol=1e-2)

            approx = reader.read_rescaled(sc, of)
            np.testing.assert_allclose(approx.ravel(), data.array.ravel(), rtol=1e-1, atol=1e-2)
        else:
            tmp = data.array
            # if the compression is None, the scale and offset should not be written to the json file
            with self.assertRaises(OSError) as context:
                sc, of = reader.read_scale_offset()

        assert tmp.dtype == read_array.dtype

        np.testing.assert_array_equal(tmp, read_array)

class TestRAW(unittest.TestCase):
    def setUp(self) -> None:
        self.TMP = tempfile.TemporaryDirectory()
        self.cwd = os.path.join(self.TMP.name, 'rawtest')

    def tearDown(self) -> None:
        self.TMP.cleanup()
        # pass

    def test_raw_nocompression_0(self):
        self.RAW_compression_test(None,1)

    def test_raw_compression_0(self):
        self.RAW_compression_test('uint8',1)

    def test_raw_compression_1(self):
        self.RAW_compression_test('uint16',1)

    def test_raw_nocompression_1(self):
        self.RAW_compression_test(None,1)

    def test_raw_compression_2(self):
        with self.assertRaises(ValueError) as context:
            self.RAW_compression_test(12,1)

    def RAW_compression_test(self, compression, channels=1):
        X=4
        Y=5
        Z=6
        C=channels
        if C == 1:
            ig = ImageGeometry(voxel_num_x=4, voxel_num_y=5, voxel_num_z=6)
        else:
            ig = ImageGeometry(voxel_num_x=4, voxel_num_y=5, voxel_num_z=6, channels=C)
        data = ig.allocate(0)
        data.fill(np.arange(X*Y*Z*C).reshape(ig.shape))

        compress = utilities.get_compress(compression)
        dtype = utilities.get_compressed_dtype(data.array, compression)
        scale, offset = utilities.get_compression_scale_offset(data.array, compression)
        if C > 1:
            assert data.ndim == 4
        raw = "unittest.raw"
        fname = os.path.join(self.cwd, raw)

        writer = RAWFileWriter(data=data, file_name=fname, compression=compression)
        writer.write()

        # read the data from the ini file
        ini = "unittest.ini"
        config = configparser.ConfigParser()
        inifname = os.path.join(self.cwd, ini)
        config.read(inifname)


        assert raw == config['MINIMAL INFO']['file_name']

        # read how to read the data from the ini file
        read_dtype = config['MINIMAL INFO']['data_type']
        read_array = np.fromfile(fname, dtype=read_dtype)
        read_shape = eval(config['MINIMAL INFO']['shape'])

        # reshape read in array
        read_array = read_array.reshape(read_shape)

        if compress:
            # rescale the dataset to the original data
            sc = float(config['COMPRESSION']['scale'])
            of = float(config['COMPRESSION']['offset'])
            assert sc == scale
            assert of == offset

            recovered_data = (read_array - of)/sc
            np.testing.assert_allclose(recovered_data, data.array, rtol=1e-1, atol=1e-2)

            # rescale the original data with scale and offset and compare with what saved
            tmp = data.array * scale + offset
            tmp = np.asarray(tmp, dtype=dtype)
        else:
            tmp = data.array

        assert tmp.dtype == read_array.dtype

        np.testing.assert_array_equal(tmp, read_array)

class Test_HDF5_utilities(unittest.TestCase):
    def setUp(self) -> None:
        data_dir = dataexample.CILDATA.data_dir
        
        self.path = os.path.join(os.path.abspath(data_dir), '24737_fd_normalised.nxs')


        self.dset_path ='/entry1/tomo_entry/data/data'


    def test_print_metadata(self):
        devnull = open(os.devnull, 'w') #suppress stdout
        with patch('sys.stdout', devnull):
            HDF5_utilities.print_metadata(self.path)


    def test_get_dataset_metadata(self):
        dset_dict = HDF5_utilities.get_dataset_metadata(self.path, self.dset_path)

        dict_by_hand  ={'ndim': 3, 'shape': (91, 135, 160), 'size': 1965600, 'dtype': np.float32, 'compression': None, 'chunks': None, 'is_virtual': False}
        self.assertEqual(dset_dict, dict_by_hand | dset_dict)


    def test_read(self):

        data_full = dataexample.SYNCHROTRON_PARALLEL_BEAM_DATA.get()

        # full dataset
        data_read_full = HDF5_utilities.read(self.path, self.dset_path)
        np.testing.assert_allclose(data_full.array,data_read_full)

        # subset of input
        subset = np.s_[44:45,70:90:2,80]
        data_read_subset = HDF5_utilities.read(self.path, self.dset_path, subset)
        self.assertTrue(data_read_subset.dtype == np.float32)
        np.testing.assert_allclose(data_full.array[subset],data_read_subset)

        # read as dtype
        subset = np.s_[44:45,70:90:2,80]
        data_read_dtype = HDF5_utilities.read(self.path, self.dset_path, subset, dtype=np.float64)
        self.assertTrue(data_read_dtype.dtype == np.float64)
        np.testing.assert_allclose(data_full.array[subset],data_read_dtype)


    def test_read_to(self):
        data_full = dataexample.SYNCHROTRON_PARALLEL_BEAM_DATA.get()

        # full dataset
        data_full_out = np.empty_like(data_full.array, dtype=np.float32)
        HDF5_utilities.read_to(self.path, self.dset_path, data_full_out)
        np.testing.assert_allclose(data_full.array,data_full_out)

        # subset of input, continuous output
        subset = np.s_[44:45,70:90:2,80]
        data_subset_out = np.empty((1,10), dtype=np.float32)
        HDF5_utilities.read_to(self.path, self.dset_path, data_subset_out, source_sel=subset)
        np.testing.assert_allclose(data_full.array[subset],data_subset_out)

        # subset of input, continuous output, change of dtype
        subset = np.s_[44:45,70:90:2,80]
        data_subset_out = np.empty((1,10), dtype=np.float64)
        HDF5_utilities.read_to(self.path, self.dset_path, data_subset_out, source_sel=subset)
        np.testing.assert_allclose(data_full.array[subset],data_subset_out)

        # subset of input written to subset of  output
        data_partial_by_hand = np.zeros_like(data_full.array, dtype=np.float32)
        data_partial = np.zeros_like(data_full.array, dtype=np.float32)

        data_partial_by_hand[subset] = data_full.array[subset]

        HDF5_utilities.read_to(self.path, self.dset_path, data_partial, source_sel=subset, dest_sel=subset)
        np.testing.assert_allclose(data_partial_by_hand,data_partial)


class TestNikonReader(unittest.TestCase):

    def test_setup(self):

        reader = NikonDataReader()
        self.assertEqual(reader.file_name, None)
        self.assertEqual(reader.roi, None)
        self.assertTrue(reader.normalise)
        self.assertEqual(reader.mode, 'bin')
        self.assertFalse(reader.fliplr)

        roi = {'vertical':(1,-1),'horizontal':(1,-1),'angle':(1,-1)}
        reader = NikonDataReader(file_name=None, roi=roi, normalise=False, mode='slice', fliplr=True)
        self.assertEqual(reader.file_name, None)
        self.assertEqual(reader.roi, roi)
        self.assertFalse(reader.normalise)
        self.assertEqual(reader.mode, 'slice')
        self.assertTrue(reader.fliplr)

        with self.assertRaises(FileNotFoundError):
            reader = NikonDataReader(file_name='no-file')



#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
# Kyle Pidgeon (UKRI-STFC)

from ipywidgets import interactive_output
import ipywidgets as widgets

import matplotlib.pyplot as plt
from matplotlib import gridspec

import numpy

from IPython.display import HTML, display
import random
from cil.utilities.display import set_origin


def display_slice(container, clim, direction, title, cmap, size, axis_labels, origin):


    def get_slice_3D(x, minmax, roi_hdir, roi_vdir, equal_aspect):

        if direction == 0:
            img = container[x]
            x_label = axis_labels[2]
            y_label = axis_labels[1]

        elif direction == 1:
            img = container[:,x,:]
            x_label = axis_labels[2]
            y_label = axis_labels[0]

        elif direction == 2:
            img = container[:,:,x]
            x_label = axis_labels[1]
            y_label = axis_labels[0]

        if size is None:
            fig = plt.figure()
        else:
            fig = plt.figure(figsize=size)

        dtitle = ''
        if isinstance(title, (list, tuple)):
            dtitle = title[x]
        else:
            dtitle = title

        gs = gridspec.GridSpec(1, 2, figure=fig, width_ratios=(1,.05), height_ratios=(1,))
        # image
        ax = fig.add_subplot(gs[0, 0])
        img, data_origin, _ = set_origin(img, origin)

        aspect = 'equal'
        if not equal_aspect:
            aspect = (roi_hdir[1] - roi_hdir[0]) / (roi_vdir[1] - roi_vdir[0])

        if 'right' in origin:
            roi_hdir = roi_hdir[1], roi_hdir[0]
        if 'upper' in origin:
            roi_vdir = roi_vdir[1], roi_vdir[0]

        aximg = ax.imshow(img, cmap=cmap, origin=data_origin, aspect=aspect)
        cmin = clim[0] + (minmax[0] / 100)*(clim[1]-clim[0])
        cmax = clim[0] + (minmax[1] / 100)*(clim[1]-clim[0])
        aximg.set_clim((cmin, cmax))
        ax.set_xlim(*roi_hdir)
        ax.set_ylim(*roi_vdir)
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.set_title(f'{dtitle} {x}')
        # colorbar
        ax = fig.add_subplot(gs[0, 1])
        plt.colorbar(aximg, cax=ax)
        plt.tight_layout()
        plt.show(fig)

    return get_slice_3D


def islicer(data, direction=0, title=None, slice_number=None, cmap='gray',
            minmax=None, size=None, axis_labels=None, origin='lower-left',
            play_interval=500):
    """
    Creates an interactive slider that slices a 3D volume along an axis.

    Parameters
    ----------
    data : DataContainer or numpy.ndarray
        A 3-dimensional dataset from which 2-dimensional slices will be
        shown
    direction : int
        Axis to slice on. Can be 0,1,2 or the axis label, default 0
    title : str, list of str or tuple of str, default=''
        Title for the display
    slice_number : int, optional
        Start slice number (default is None, which results in the center
        slice being shown initially)
    cmap : str or matplotlib.colors.Colormap, default='gray'
        Set the colour map
    minmax : tuple
        Colorbar (min, max) values, default None (uses the min, max of
        values in `data`)
    size : int or tuple, optional
        Specify the figure size in inches. If `int` this specifies the
        width, and scales the height in order to keep the standard
        `matplotlib` aspect ratio, default None (use the default matplotlib
        figure size)
    axis_labels : list of str, optional
        The axis labels to use for each of the 3 dimensions in the data
        (default is None, resulting in labels extracted from the data, or
        ['X','Y','Z'] if no labels are present)
    origin : {'lower-left', 'upper-left', 'lower-right', 'upper-right'}
        Sets the display origin
    play_interval : int, default=500
        The interval of time (in ms) a slice is selected for when iterating
        through a set of them

    Returns
    -------
    box : ipywidgets.Box
        The top-level widget container.
    """

    if axis_labels is None:
        if hasattr(data, "dimension_labels"):
            axis_labels = [*data.dimension_labels]
        else:
            axis_labels = ['X', 'Y', 'Z']

    if isinstance (data, numpy.ndarray):
        container = data
    elif hasattr(data, "__getitem__"):
        container = data
    elif hasattr(data, "as_array"):
        container = data.as_array()

    if not isinstance (direction, int):
        if direction in data.dimension_labels:
            direction = data.get_dimension_axis(direction)

    if slice_number is None:
        slice_number = int(data.shape[direction]/2)

    if title is None:
        title = "Direction {}: Slice".format(axis_labels[direction])

    style = {'slider_width': '80%'}
    layout = widgets.Layout(width='200px')

    slice_slider = widgets.IntSlider(
        min=0,
        max=data.shape[direction]-1,
        step=1,
        value=slice_number,
        continuous_update=True,
        layout=layout,
        style=style,
    )
    slice_selector_full = widgets.VBox([widgets.Label('Slice index (direction {})'.format(axis_labels[direction])), slice_slider])


    play_slices = widgets.Play(
        min=0,
        max=data.shape[direction]-1,
        step=1,
        interval=play_interval,
        value=slice_number,
        disabled=False,
    )
    widgets.jslink((play_slices, 'value'), (slice_slider, 'value'))

    amax = container.max()
    amin = container.min()
    if minmax is None:
        minmax = (amin, amax)

    if isinstance (size, (int, float)):
        default_ratio = 6./8.
        size = ( size , size * default_ratio )

    min_max = widgets.IntRangeSlider(
        value=[0, 100],
        min=0,
        max=100,
        step=5,
        disabled=False,
        continuous_update=True,
        orientation='horizontal',
        readout=True,
        layout=layout,
        style=style,
    )
    min_max_full = widgets.VBox([widgets.Label('Display window percent'), min_max])

    dirs_remaining = [i for i in range(3) if i != direction]
    h_dir, v_dir = dirs_remaining[1], dirs_remaining[0]
    h_dir_size = container.shape[h_dir]
    v_dir_size = container.shape[v_dir]

    roi_select_hdir = widgets.IntRangeSlider(
        value=[0, h_dir_size-1],
        min=0,
        max=h_dir_size-1,
        step=1,
        disabled=False,
        continuous_update=False,
        orientation='horizontal',
        readout=True,
        readout_format='d',
        layout=layout,
        style=style,
    )
    roi_select_hdir_full = widgets.VBox([widgets.Label(f'Range: {axis_labels[h_dir]}'), roi_select_hdir])


    roi_select_vdir = widgets.IntRangeSlider(
        value=[0, v_dir_size-1],
        min=0,
        max=v_dir_size-1,
        step=1,
        disabled=False,
        continuous_update=False,
        orientation='horizontal',
        readout=True,
        readout_format='d',
        layout=layout,
        style=style,
    )
    roi_select_vdir_full = widgets.VBox([widgets.Label(f'Range: {axis_labels[v_dir]}'), roi_select_vdir])

    equal_aspect = widgets.Checkbox(
        value=True,
        description='Pixel aspect ratio = 1',
        disabled=False,
        indent=False,
        layout=widgets.Layout(width='auto'),
    )

    box_layout = widgets.Layout(
        display='flex',
        flex_flow='column',
        align_items='flex-start',
        justify_content='center',
    )
    selectors = widgets.Box([
        play_slices,
        slice_selector_full,
        min_max_full,
        roi_select_hdir_full,
        roi_select_vdir_full,
        equal_aspect],
        layout=box_layout)

    out = interactive_output(
        display_slice(
            container,
            minmax,
            direction,
            title=title,
            cmap=cmap,
            size=size,
            axis_labels=axis_labels,
            origin=origin),
        {'x': slice_slider,
        'minmax': min_max,
        'roi_hdir': roi_select_hdir,
        'roi_vdir': roi_select_vdir,
        'equal_aspect': equal_aspect})

    box = widgets.HBox(children=[out, selectors],
                       layout=widgets.Layout(
                            display='flex',
                            justify_content='center'))

    return box


def link_islicer(*args):
    '''Links islicer's slice-selection widgets

    Parameters
    ----------
    *args : tuple of ipywidgets.Box
        The widget containers returned from `islicer`, from which the slice
        selection sliders will be extracted and linked.
    '''
    slice_sliders = [arg.children[-1].children[1].children[1] for arg in args]
    play_widgets = [arg.children[-1].children[0] for arg in args][1:]
    for p in play_widgets:
        p.layout.visibility = 'hidden'
    linked = [(widg, 'value') for widg in slice_sliders]
    # link pair-wise
    pairs = [(linked[i+1],linked[i]) for i in range(len(linked)-1)]
    for pair in pairs:
        widgets.link(*pair)

    display(*args)

# https://stackoverflow.com/questions/31517194/how-to-hide-one-specific-cell-input-or-output-in-ipython-notebook/52664156

def hide_toggle(for_next=False):
    this_cell = """$('div.cell.code_cell.rendered.selected')"""
    next_cell = this_cell + '.next()'

    toggle_text = 'Toggle show/hide'  # text shown on toggle link
    target_cell = this_cell  # target cell to control with toggle
    js_hide_current = ''  # bit of JS to permanently hide code in current cell (only when toggling next cell)

    if for_next:
        target_cell = next_cell
        toggle_text += ' next cell'
        js_hide_current = this_cell + '.find("div.input").hide();'

    js_f_name = 'code_toggle_{}'.format(str(random.randint(1,2**64)))

    html = """
        <script>
            function {f_name}() {{
                {cell_selector}.find('div.input').toggle();
            }}

            {js_hide_current}
        </script>

        <a href="javascript:{f_name}()">{toggle_text}</a>
    """.format(
        f_name=js_f_name,
        cell_selector=target_cell,
        js_hide_current=js_hide_current,
        toggle_text=toggle_text
    )

    return HTML(html)



#  Copyright 2019 United Kingdom Research and Innovation
#  Copyright 2019 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

import unittest
from utils import initialise_tests
import os
import sys
from testclass import CCPiTestClass
import numpy as np
from cil.optimisation.utilities import Sampler, SamplerRandom
initialise_tests()

sys.path.append(os.path.dirname(os.path.abspath(__file__)))


class TestSamplers(CCPiTestClass):

    def example_function(self, iteration_number):
        return ((iteration_number+5) % 50)

    def test_init_Sampler(self):
        default_sampler = Sampler(50, self.example_function)
        self.assertEqual(default_sampler._num_indices, 50)
        self.assertEqual(default_sampler.num_indices, 50)
        self.assertListEqual(default_sampler._prob_weights, [1/50]*50)
        self.assertListEqual(default_sampler.prob_weights, [1/50]*50)
        self.assertEqual(default_sampler._iteration_number, 0)
        self.assertEqual(default_sampler.current_iter_number, 0)
        self.assertEqual(default_sampler._type, None)

        other_sampler = Sampler(55, self.example_function, sampling_type='banana',
                                prob_weights=list(range(55)/np.sum(range(55))))
        self.assertEqual(other_sampler._num_indices, 55)
        self.assertEqual(other_sampler.num_indices, 55)
        self.assertListEqual(other_sampler._prob_weights,
                             list(range(55)/np.sum(range(55))))
        self.assertListEqual(other_sampler.prob_weights,
                             list(range(55)/np.sum(range(55))))
        self.assertEqual(other_sampler._iteration_number, 0)
        self.assertEqual(other_sampler.current_iter_number, 0)
        self.assertEqual(other_sampler._type, 'banana')

        # Check probabilities sum to one and are positive
        with self.assertRaises(ValueError):
            Sampler.from_function(
                10, self.example_function, prob_weights=[1/11]*10)
        with self.assertRaises(ValueError):
            Sampler.from_function(10, self.example_function,
                                  prob_weights=[-1]+[2]+[0]*8)

        # Check function is callable
        with self.assertRaises(ValueError):
            Sampler.from_function(10, function='banana')

        # check num_subset is an integer
        with self.assertRaises(ValueError):
            Sampler.from_function(10.5, self.example_function)

    def test_init_RandomSampler(self):
        default_sampler = SamplerRandom(10)
        self.assertEqual(default_sampler._num_indices, 10)
        self.assertEqual(default_sampler.num_indices, 10)
        self.assertListEqual(default_sampler._prob_weights, [1/10]*10)
        self.assertListEqual(default_sampler.prob_weights, [1/10]*10)
        self.assertEqual(default_sampler._replace, True)
        self.assertEqual(default_sampler.replace, True)
        self.assertEqual(default_sampler._iteration_number, 0)
        self.assertEqual(default_sampler.current_iter_number, 0)
        self.assertEqual(default_sampler._type, 'random_with_replacement')
        self.assertEqual(default_sampler._sampling_list, None)

        other_sampler = SamplerRandom(
            11, seed=3, sampling_type='banana', replace=False, prob=list(range(11)/np.sum(range(11))))
        self.assertEqual(other_sampler._num_indices, 11)
        self.assertEqual(other_sampler.num_indices, 11)
        self.assertEqual(other_sampler._seed, 3)
        self.assertEqual(other_sampler.seed, 3)
        self.assertEqual(other_sampler._replace, False)
        self.assertEqual(other_sampler.replace, False)
        self.assertListEqual(other_sampler._prob_weights,
                             list(range(11)/np.sum(range(11))))
        self.assertListEqual(other_sampler.prob_weights,
                             list(range(11)/np.sum(range(11))))
        self.assertEqual(other_sampler._iteration_number, 0)
        self.assertEqual(other_sampler.current_iter_number, 0)
        self.assertEqual(other_sampler._type, 'banana')
        self.assertEqual(other_sampler._sampling_list, None)

    def test_from_function(self):

        sampler = Sampler.from_function(50, self.example_function)
        order = [5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18,
                 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32]
        self.assertNumpyArrayEqual(sampler.get_previous_samples(), np.array([]))
        self.assertNumpyArrayEqual(sampler.view_samples(20), np.array(
            order)[:20])

        N = 25
        for i in range(N):
            self.assertEqual(next(sampler), order[i])
            self.assertEqual(sampler.get_current_sample(), order[i])

        self.assertEqual(sampler._iteration_number, N)
        self.assertEqual(sampler.current_iter_number, N)

        self.assertEqual(sampler.view_samples(
            550)[519], self.example_function(519))
        
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(sampler.get_samples(
            550)[519], self.example_function(519))
            
        
        self.assertNumpyArrayEqual(sampler.get_previous_samples(), np.array(order[:N]))

        sampler = Sampler.from_function(50, self.example_function)
        self.assertListEqual(sampler.prob_weights, [1/50] * 50)
        self.assertEqual(sampler.num_indices, 50)
        self.assertEqual(sampler._type, 'from_function')

        sampler = Sampler.from_function(40, self.example_function, [1]+[0]*39)
        self.assertListEqual(sampler.prob_weights,  [1]+[0]*39)
        self.assertEqual(sampler.num_indices, 40)
        self.assertEqual(sampler._type, 'from_function')
        
    def test_sequential_iterator_and_view_samples(self):

        sampler = Sampler.sequential(10)
        self.assertEqual(sampler.num_indices, 10)
        self.assertEqual(sampler._type, 'sequential')
        self.assertListEqual(sampler.prob_weights, [1/10]*10)

        sampler = Sampler.sequential(10)
        self.assertNumpyArrayEqual(sampler.view_samples(20), np.array(
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9]))
        
        with self.assertRaises(ValueError):
            sampler.get_current_sample()

        for i in range(337):
            self.assertEqual(next(sampler), i % 10)
            self.assertEqual(sampler.get_current_sample(), i % 10)
            self.assertEqual(sampler.get_current_sample(), i % 10)
            
        self.assertEqual(sampler.current_iter_number, 337)
        self.assertEqual(len(sampler.get_previous_samples()), 337)

        self.assertNumpyArrayEqual(sampler.view_samples(20), np.array(
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9]))

    def test_random_without_replacement_iterator_and_view_samples(self):

        sampler = Sampler.random_without_replacement(8, seed=1)
        self.assertEqual(sampler.num_indices, 8)
        self.assertEqual(sampler._type, 'random_without_replacement')
        self.assertEqual(sampler._prob_weights,  [1/8]*8)
        self.assertEqual(sampler._seed, 1)
        self.assertListEqual(sampler.prob_weights, sampler._prob_weights)

        sampler = Sampler.random_without_replacement(7)
        self.assertEqual(sampler.num_indices, 7)
        self.assertEqual(sampler._type, 'random_without_replacement')
        self.assertEqual(sampler._prob_weights, [1/7]*7)
        self.assertListEqual(sampler.prob_weights, sampler._prob_weights)

        sampler = Sampler.random_without_replacement(7, seed=1)
        order = [2, 5, 0, 1, 4, 3, 6, 1, 6, 0, 4, 2, 3, 5, 5, 6, 2, 4, 0, 1, 3, 0,
                 2, 6, 3]
        self.assertNumpyArrayEqual(
            sampler.view_samples(25), np.array(order[:25]))

        with self.assertRaises(ValueError):
            sampler.get_current_sample()
            
        for i in range(25):
            self.assertEqual(next(sampler), order[i])
            self.assertEqual(sampler.get_current_sample(), order[i])

        self.assertNumpyArrayEqual(
            sampler.view_samples(25), np.array(order[:25]))
        
        with self.assertWarns(DeprecationWarning):
            self.assertNumpyArrayEqual(
            sampler.get_samples(25), np.array(order[:25]))

    def test_herman_meyer_iterator_and_view_samples(self):

        sampler = Sampler.herman_meyer(12)
        self.assertEqual(sampler.num_indices, 12)
        self.assertEqual(sampler._type, 'herman_meyer')
        self.assertListEqual(sampler.prob_weights, [1/12] * 12)
        out = [sampler.next() for _ in range(12)]
        self.assertListEqual(out, [0, 6, 3, 9, 1, 7, 4, 10, 2, 8, 5, 11])

        sampler = Sampler.herman_meyer(12)
        order = [0, 6, 3, 9, 1, 7, 4, 10, 2, 8, 5,
                 11, 0, 6, 3, 9, 1, 7, 4, 10, 2, 8, 5, 11]
        self.assertNumpyArrayEqual(
            sampler.view_samples(14), np.array(order[:14]))

        for i in range(25):
            self.assertEqual(sampler.current_iter_number, i)
            self.assertEqual(sampler.next(), order[i % 12])
            self.assertEqual(sampler.get_current_sample(), order[i % 12])
            self.assertEqual(sampler.get_current_sample(), order[i % 12])
            
        self.assertNumpyArrayEqual(sampler.view_samples(25), sampler.get_previous_samples())

        self.assertNumpyArrayEqual(
            sampler.view_samples(14), np.array(order[:14]))

    def test_random_with_replacement_iterator_and_view_samples(self):

        sampler = Sampler.random_with_replacement(5)
        self.assertEqual(sampler.num_indices, 5)
        self.assertEqual(sampler._type, 'random_with_replacement')
        self.assertListEqual(sampler._prob_weights, [1/5] * 5)
        self.assertListEqual(sampler.prob_weights, [1/5] * 5)

        sampler = Sampler.random_with_replacement(4, [0.7, 0.1, 0.1, 0.1])
        self.assertEqual(sampler.num_indices, 4)
        self.assertEqual(sampler._type, 'random_with_replacement')
        self.assertListEqual(sampler.prob_weights, [0.7, 0.1, 0.1, 0.1])

        sampler = Sampler.random_with_replacement(5, seed=5)
        order = [1, 4, 1, 4, 2, 3, 3, 2, 1, 0, 0, 3,
                 2, 0, 4, 1, 2, 1, 3, 2, 2, 1, 1, 1, 1]
        self.assertNumpyArrayEqual(
            sampler.view_samples(14), np.array(order[:14]))
        
        with self.assertRaises(ValueError):
            sampler.get_current_sample()

        for i in range(25):
            self.assertEqual(next(sampler), order[i])
            self.assertEqual(sampler.get_current_sample(), order[i])
            self.assertEqual(sampler.get_current_sample(), order[i])
            self.assertEqual(sampler.current_iter_number, i+1)

        self.assertNumpyArrayEqual(
            sampler.view_samples(14), np.array(order[:14]))

        self.assertNumpyArrayEqual(sampler.view_samples(25), sampler.get_previous_samples())
        
        sampler = Sampler.random_with_replacement(
            4, [0.7, 0.1, 0.1, 0.1], seed=5)
        order = [0, 2, 0, 3, 0, 0, 1, 0, 0, 0, 0, 1,
                 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.assertNumpyArrayEqual(
            sampler.view_samples(14), np.array(order[:14]))

        for i in range(25):
            self.assertEqual(sampler.next(), order[i])
            self.assertEqual(sampler.get_current_sample(), order[i])
            self.assertEqual(sampler.get_current_sample(), order[i])
            self.assertEqual(sampler.current_iter_number, i+1)
            
        self.assertNumpyArrayEqual(sampler.view_samples(25), sampler.get_previous_samples())

        self.assertNumpyArrayEqual(
            sampler.view_samples(14), np.array(order[:14]))

    def test_staggered_iterator_and_view_samples(self):

        sampler = Sampler.staggered(21, 4)
        self.assertEqual(sampler.num_indices, 21)
        self.assertEqual(sampler._type, 'staggered')
        out = [sampler.next() for _ in range(21)]
        self.assertListEqual(
            out, [0, 4, 8, 12, 16, 20, 1, 5, 9, 13, 17, 2, 6, 10, 14, 18, 3, 7, 11, 15, 19])
        self.assertListEqual(sampler.prob_weights, [1/21] * 21)

        with self.assertRaises(ValueError):
            Sampler.staggered(22, 25)

        sampler = Sampler.staggered(21, 4)
        order = [0, 4, 8, 12, 16, 20, 1, 5, 9, 13,
                 17, 2, 6, 10, 14, 18, 3, 7, 11, 15, 19]
        self.assertNumpyArrayEqual(
            sampler.view_samples(10), np.array(order[:10]))

        for i in range(25):
            self.assertEqual(next(sampler), order[i % 21])
            self.assertEqual(sampler.get_current_sample(), order[i % 21])
            self.assertEqual(sampler.get_current_sample(), order[i % 21])
            self.assertEqual(sampler.current_iter_number, i+1)

        self.assertNumpyArrayEqual(
            sampler.view_samples(10), np.array(order[:10]))
        
        self.assertNumpyArrayEqual(sampler.view_samples(25), sampler.get_previous_samples())




#  Copyright 2018 United Kingdom Research and Innovation
#  Copyright 2018 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
# Joshua DM Hellier (University of Manchester) [refactorer]
import math

import numpy

from .block import BlockGeometry


class Partitioner(object):
    '''Interface for AcquisitionData to be able to partition itself in a number of batches.

    This class, by multiple inheritance with AcquisitionData, allows the user to partition the data,
    by using the method ``partition``.
    The partitioning will generate a ``BlockDataContainer`` with appropriate ``AcquisitionData``.

    '''
    # modes of partitioning
    SEQUENTIAL = 'sequential'
    STAGGERED = 'staggered'
    RANDOM_PERMUTATION = 'random_permutation'

    def _partition_indices(self, num_batches, indices, stagger=False):
        """Partition a list of indices into num_batches of indices.

        Parameters
        ----------
        num_batches : int
            The number of batches to partition the indices into.
        indices : list of int, int
            The indices to partition. If passed a list, this list will be partitioned in ``num_batches``
            partitions. If passed an int the indices will be generated automatically using ``range(indices)``.
        stagger : bool, default False
            If True, the indices will be staggered across the batches.

        Returns
        --------
        list of list of int
            A list of batches of indices.
        """

        # Partition the indices into batches.
        if isinstance(indices, int):
            indices = list(range(indices))

        num_indices = len(indices)
        # sanity check
        if num_indices < num_batches:
            raise ValueError(
                'The number of batches must be less than or equal to the number of indices.'
            )

        if stagger:
            batches = [indices[i::num_batches] for i in range(num_batches)]

        else:
            # we split the indices with floor(N/M)+1 indices in N%M groups
            # and floor(N/M) indices in the remaining M - N%M groups.

            # rename num_indices to N for brevity
            N = num_indices
            # rename num_batches to M for brevity
            M = num_batches
            batches = [
                indices[j:j + math.floor(N / M) + 1] for j in range(N % M)
            ]
            offset = N % M * (math.floor(N / M) + 1)
            for i in range(M - N % M):
                start = offset + i * math.floor(N / M)
                end = start + math.floor(N / M)
                batches.append(indices[start:end])

        return batches

    def _construct_BlockGeometry_from_indices(self, indices):
        '''Convert a list of boolean masks to a list of BlockGeometry.

        Parameters
        ----------
          indices : list of lists of indices

        Returns
        -------
            BlockGeometry
        '''
        ags = []
        for mask in indices:
            ag = self.geometry.copy()
            ag.config.angles.angle_data = numpy.take(self.geometry.angles, mask, axis=0)
            ags.append(ag)
        return BlockGeometry(*ags)

    def partition(self, num_batches, mode, seed=None):
        '''Partition the data into ``num_batches`` batches using the specified ``mode``.
        
        The modes are

        1. ``sequential`` - The data will be partitioned into ``num_batches`` batches of sequential indices.

        2. ``staggered`` - The data will be partitioned into ``num_batches`` batches of sequential indices, with stride equal to ``num_batches``.

        3. ``random_permutation`` - The data will be partitioned into ``num_batches`` batches of random indices.

        Parameters
        ----------
        num_batches : int
            The number of batches to partition the data into.
        mode : str
            The mode to use for partitioning. Must be one of ``sequential``, ``staggered`` or ``random_permutation``.
        seed : int, optional
            The seed to use for the random permutation. If not specified, the random number
            generator will not be seeded.


        Returns
        -------
        BlockDataContainer
            Block of `AcquisitionData` objects containing the data requested in each batch

        Note
        ----
        This only works on datasets with an 'angle' dimension, and is not currently implemented for Cone3D_Flex geometry.

        Example
        -------

        Partitioning a list of ints [0, 1, 2, 3, 4, 5, 6, 7, 8] into 4 batches will return:

        1. [[0, 1, 2], [3, 4], [5, 6], [7, 8]] with ``sequential``
        2. [[0, 4, 8], [1, 5], [2, 6], [3, 7]] with ``staggered``
        3. [[8, 2, 6], [7, 1], [0, 4], [3, 5]] with ``random_permutation`` and seed 1

        '''
        if 'angle' not in self.dimension_labels:
            raise NotImplementedError(f"Currently, Partitioner only partitions on the `angle` dimension \
                and therefore can only be used on datasets with an 'angle' dimension. \
                Dimensions provided: {self.dimension_labels}")
        
        if mode == Partitioner.SEQUENTIAL:
            return self._partition_deterministic(num_batches, stagger=False)
        elif mode == Partitioner.STAGGERED:
            return self._partition_deterministic(num_batches, stagger=True)
        elif mode == Partitioner.RANDOM_PERMUTATION:
            return self._partition_random_permutation(num_batches, seed=seed)
        else:
            raise ValueError('Unknown partition mode {}'.format(mode))

    def _partition_deterministic(self, num_batches, stagger=False, indices=None):
        '''Partition the data into ``num_batches`` batches.

        Parameters
        ----------
        num_batches : int
            The number of batches to partition the data into.
        stagger : bool, optional
            If ``True``, the batches will be staggered. Default is ``False``.
        indices : list of int, optional
            The indices to partition. If not specified, the indices will be generated from the number of projections.
        '''
        if indices is None:
            indices = self.geometry.num_projections
        partition_indices = self._partition_indices(num_batches, indices, stagger)
        blk_geo = self._construct_BlockGeometry_from_indices(partition_indices)

        # copy data
        out = blk_geo.allocate(None)
        axis = self.dimension_labels.index('angle')

        for i in range(num_batches):
            out[i].fill(
                numpy.squeeze(
                    numpy.take(self.array, partition_indices[i], axis=axis)
                )
            )

        return out

    def _partition_random_permutation(self, num_batches, seed=None):
        '''Partition the data into ``num_batches`` batches using a random permutation.

        Parameters
        ----------
        num_batches : int
            The number of batches to partition the data into.
        seed : int, optional
            The seed to use for the random permutation. If not specified, the random number generator
            will not be seeded.

        '''
        if seed is not None:
            numpy.random.seed(seed)

        indices = numpy.arange(self.geometry.num_projections)
        numpy.random.shuffle(indices)

        indices = list(indices)

        return self._partition_deterministic(num_batches, stagger=False, indices=indices)



#  Copyright 2021 United Kingdom Research and Innovation
#  Copyright 2021 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
import logging

import numpy as np

from cil.framework import ImageData, AcquisitionData, BlockGeometry
from cil.framework.labels import AcquisitionDimension, ImageDimension
from cil.optimisation.operators import BlockOperator, LinearOperator
from cil.plugins.tigre import CIL2TIGREGeometry

log = logging.getLogger(__name__)

try:
    from _Atb import _Atb_ext as Atb
    from _Ax import _Ax_ext as Ax

except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "This plugin requires the additional package TIGRE\n" +
        "Please install it via conda as tigre from the ccpi channel")

try:
    from tigre.utilities.gpu import GpuIds
    has_gpu_sel = True
except ModuleNotFoundError:
    has_gpu_sel = False


class ProjectionOperator(LinearOperator):
    """
        ProjectionOperator configures and calls TIGRE Projectors for your dataset.

        Please refer to the TIGRE documentation for further descriptions
        https://github.com/CERN/TIGRE
        https://iopscience.iop.org/article/10.1088/2057-1976/2/5/055010


        Parameters
        ----------

        image_geometry : `ImageGeometry`, default used if None
            A description of the area/volume to reconstruct

        acquisition_geometry :`AcquisitionGeometry`, `BlockGeometry`
            A description of the acquisition data. If passed a BlockGeometry it will return a BlockOperator.

        direct_method : str,  default 'interpolated'
            The method used by the forward projector, 'Siddon' for ray-voxel intersection, 'interpolated' for interpolated projection

        adjoint_weights : str, default 'matched'
            The weighting method used by the cone-beam backward projector, 'matched' for weights to approximately match the 'interpolated' forward projector, 'FDK' for FDK weights

        Example
        -------
        >>> from cil.plugins.tigre import ProjectionOperator
        >>> PO = ProjectionOperator(image.geometry, data.geometry)
        >>> forward_projection = PO.direct(image)
        >>> backward_projection = PO.adjoint(data)

        """
    def __new__(cls, image_geometry=None, acquisition_geometry=None, \
        direct_method='interpolated',adjoint_weights='matched', **kwargs):
        if isinstance(acquisition_geometry, BlockGeometry):
            log.info("BlockOperator is returned.")

            K = []
            for ag in acquisition_geometry:
                K.append(
                    ProjectionOperator_ag(image_geometry=image_geometry, acquisition_geometry=ag, \
                        direct_method=direct_method, adjoint_weights=adjoint_weights, **kwargs)
                )
            return BlockOperator(*K)
        else:
            log.info("Standard Operator is returned.")
            return super(ProjectionOperator,
                         cls).__new__(ProjectionOperator_ag)


class ProjectionOperator_ag(ProjectionOperator):
    '''TIGRE Projection Operator'''

    def __init__(self,
                 image_geometry=None,
                 acquisition_geometry=None,
                 direct_method='interpolated',
                 adjoint_weights='matched',
                 **kwargs):
        """
        ProjectionOperator configures and calls TIGRE Projectors for your dataset.

        Please refer to the TIGRE documentation for further descriptions
        https://github.com/CERN/TIGRE
        https://iopscience.iop.org/article/10.1088/2057-1976/2/5/055010


        Parameters
        ----------

        image_geometry : ImageGeometry, default used if None
            A description of the area/volume to reconstruct

        acquisition_geometry : AcquisitionGeometry
            A description of the acquisition data

        direct_method : str,  default 'interpolated'
            The method used by the forward projector, 'Siddon' for ray-voxel intersection, 'interpolated' for interpolated projection

        adjoint_weights : str, default 'matched'
            The weighting method used by the cone-beam backward projector, 'matched' for weights to approximately match the 'interpolated' forward projector, 'FDK' for FDK weights

        Example
        -------
        >>> from cil.plugins.tigre import ProjectionOperator
        >>> PO = ProjectionOperator(image.geometry, data.geometry)
        >>> forward_projection = PO.direct(image)
        >>> backward_projection = PO.adjoint(data)

        """

        if acquisition_geometry is None:
            raise TypeError("Please specify an acquisition_geometry to configure this operator")

        if image_geometry == None:
            image_geometry = acquisition_geometry.get_ImageGeometry()

        device = kwargs.get('device', 'gpu')
        if device != 'gpu':
            raise ValueError(
                "TIGRE projectors are GPU only. Got device = {}".format(
                    device))

        ImageDimension.check_order_for_engine('tigre', image_geometry)
        AcquisitionDimension.check_order_for_engine('tigre', acquisition_geometry)

        super(ProjectionOperator,self).__init__(domain_geometry=image_geometry,\
             range_geometry=acquisition_geometry)

        if direct_method not in ['interpolated', 'Siddon']:
            raise ValueError(
                "direct_method expected 'interpolated' or 'Siddon' got {}".
                format(direct_method))

        if adjoint_weights not in ['matched', 'FDK']:
            raise ValueError(
                "adjoint_weights expected 'matched' or 'FDK' got {}".format(
                    adjoint_weights))

        self.method = {'direct': direct_method, 'adjoint': adjoint_weights}

        #set up TIGRE geometry
        tigre_geom, tigre_angles = CIL2TIGREGeometry.getTIGREGeometry(
            image_geometry, acquisition_geometry)

        tigre_geom.check_geo(tigre_angles)
        tigre_geom.cast_to_single()
        self.tigre_geom = tigre_geom

        #set up TIGRE GPU targets (from 2.2)
        if has_gpu_sel:
            self.gpuids = GpuIds()

    def __call_Ax(self, data):
        if has_gpu_sel:
            return Ax(data, self.tigre_geom, self.tigre_geom.angles,
                      self.method['direct'], self.tigre_geom.mode, self.gpuids)
        else:
            return Ax(data, self.tigre_geom, self.tigre_geom.angles,
                      self.method['direct'], self.tigre_geom.mode)

    def direct(self, x, out=None):

        data = x.as_array()

        if self.tigre_geom.is2D:
            data_temp = np.expand_dims(data, axis=0)
            arr_out = self.__call_Ax(data_temp)
            arr_out = np.squeeze(arr_out, axis=1)
        else:
            arr_out = self.__call_Ax(data)

        #if single angle projection remove the dimension for CIL
        if arr_out.shape[0] == 1:
            arr_out = np.squeeze(arr_out, axis=0)

        if out is None:
            out = AcquisitionData(arr_out,
                                  deep_copy=False,
                                  geometry=self._range_geometry.copy())
        else:
            out.fill(arr_out)
        return out

    def __call_Atb(self, data):
        if has_gpu_sel:
            return Atb(data, self.tigre_geom, self.tigre_geom.angles,
                       self.method['adjoint'], self.tigre_geom.mode,
                       self.gpuids)
        else:
            return Atb(data, self.tigre_geom, self.tigre_geom.angles,
                       self.method['adjoint'], self.tigre_geom.mode)

    def adjoint(self, x, out=None):

        data = x.as_array()

        #if single angle projection add the dimension in for TIGRE
        if x.dimension_labels[0] != AcquisitionDimension.ANGLE:
            data = np.expand_dims(data, axis=0)

        if self.tigre_geom.is2D:
            data = np.expand_dims(data, axis=1)
            arr_out = self.__call_Atb(data)
            arr_out = np.squeeze(arr_out, axis=0)
        else:
            arr_out = self.__call_Atb(data)

        if out is None:
            out = ImageData(arr_out,
                            deep_copy=False,
                            geometry=self._domain_geometry.copy())
        else:
            out.fill(arr_out)
        return out

    def domain_geometry(self):
        return self._domain_geometry

    def range_geometry(self):
        return self._range_geometry



#  Copyright 2025 United Kingdom Research and Innovation
#  Copyright 2025 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from abc import ABC, abstractmethod
from cil.optimisation.algorithms import Algorithm
from cil.optimisation.functions import ZeroFunction
from cil.optimisation.utilities import ConstantStepSize, StepSizeRule
import numpy
import logging
from numbers import Real, Number
import warnings

log = logging.getLogger(__name__)


class ScalarMomentumCoefficient(ABC):
    r'''Abstract base class for MomentumCoefficient objects. The `__call__` method of this class returns the momentum coefficient for the given iteration.

    The call method of the ScalarMomentumCoefficient returns a scalar value. Given access to the algorithm object, the momentum coefficient can be a function of the algorithm state.

    The `apply_momentum_in_APGD` function,  updates the solution in the APGD algorithm as 
    
    .. math:: y_{k+1}=x_{k+1}+M(x_{k+1}-x_{k}),
    
    where :math:`M` is the calculated scalar momentum value. 


    '''

    def __init__(self):
        r'''Initialises the momentum coefficient object.
        '''
        pass

    @abstractmethod
    def __call__(self, algorithm):
        r'''Returns the momentum coefficient for the given iteration.

        Parameters
        ----------
        algorithm: CIL Algorithm
            The algorithm object.
        '''

        pass

    def apply_momentum_in_APGD(self, algorithm, out=None):
        r'''Calculates the momentum cofficient, applies a scalar momentum update in the APGD algorithm and returns the next iterate.
        
        Uses the calculation, :math:`y_{k+1}=x_{k+1}+M(x_{k+1}-x_{k})`, where :math:`M` is the calculated scalar momentum value. 

        Parameters
        ----------
        algorithm: instantiated CIL Algorithm
            The algorithm object.
        out: DataContainer, default is None
            Object to contain the next iterate. 
        '''
        momentum = self.__call__(algorithm)
        return algorithm.y.sapyb(momentum, algorithm.x, 1.0, out=out)


class ConstantMomentum(ScalarMomentumCoefficient):

    '''MomentumCoefficient object that returns a constant momentum coefficient.

    Parameters
    ----------
    momentum: float
        The constant momentum coefficient.
    '''

    def __init__(self, momentum):
        self.momentum = momentum

    def __call__(self, algorithm):
        return self.momentum


class NesterovMomentum(ScalarMomentumCoefficient):
    '''MomentumCoefficient object that returns the Nesterov momentum coefficient.

    Starting with :math:`t=1`, the Nesterov algorithm updates with each iteration:

    .. math:: t_{k+1}=\dfrac{1}{2}(1+\sqrt{1+4t_{k}^2})
    
    The momentum coefficient is then returned as :math:`\dfrac{t_{k}-1}{t_{k}}`.
    '''

    def __init__(self):
        self.t = 1

    def __call__(self, algorithm):
        self.t_old = self.t
        self.t = 0.5*(1 + numpy.sqrt(1 + 4*(self.t_old**2)))
        return (self.t_old-1)/self.t


class APGD(Algorithm):

    r""" The Accelerated Proximal Gradient Descent (APGD) algorithm is used to solve:

    .. math:: \min_{x} f(x) + g(x)

    where :math:`f` is differentiable, and :math:`g` has a *simple* proximal operator.

    In each update, the algorithm computes:

    .. math:: x_{k+1} = \mathrm{prox}_{\alpha g}(y_{k} - \alpha\nabla f(y_{k}))

    where :math:`\alpha` is the :code:`step_size`.

    Then, :math:`y_{k+1}`, is then calculated from :math:`x_{k+1}`, based on a momentum rule. Note that :math:`y_0=x_0`. Users have flexibility to do this however they wish by passing to `momentum` a class that has an `apply_montemum_in_APGD` function which takes an intialised algorithm and returns the next iterate. 
    

    Currently, we have implemented options for a scalar momentum coefficient (see :class:`cil.optimisation.algorithms.APGD.ScalarMomentumCoefficient` class.). In this case, the momentum term is added as follows:

    .. math:: y_{k+1} = x_{k+1} + M(x_{k+1} - x_{k}). 

    The default momentum coefficient is the Nesterov momentum coefficient which varies with each iteration. 


    Parameters
    ----------
    initial : DataContainer
              Initial guess of ISTA. :math:`x_{0}`
    f : Function
        Differentiable function. If `None` is passed, the algorithm will use the ZeroFunction.
    g : Function or `None`
        Convex function with *simple* proximal operator. If `None` is passed, the algorithm will use the ZeroFunction.
    step_size : positive :obj:`float` or child class of :meth:`cil.optimisation.utilities.StepSizeRule`',  the default :code:`step_size` is a constant :math:`\frac{1}{L}` or 1 if `f=None`.
                Step size for the gradient step of APGD. If a float is passed, this is used as a constant step size.  If a child class of :meth:`cil.optimisation.utilities.StepSizeRule` is passed then its method :meth:`get_step_size` is called for each update. 
    preconditioner: class with an `apply` method or a function that takes an initialised CIL algorithm as an argument and modifies a provided `gradient`.
            This could be a custom `preconditioner` or one provided in :meth:`~cil.optimisation.utilities.preconditoner`. If None is passed then `self.gradient_update` will remain unmodified. 
    momentum : float or class with an `apply_momentum_in_APGD` function that takes an intialised CIL algorithm and returns the next iterate, default is Nesterov Momentum. 
            This could be a custom momentum class or one provided by CIL (see :class:`cil.optimisation.algorithms.APGD.ScalarMomentumCoefficient`). If a float is passed, scalar momentum is used with the float as a fixed constant. 


    kwargs: Keyword arguments
        Arguments from the base class :class:`.Algorithm`.


    Note
    ----
    The APGD algorithm with (default) Nesterov momentum is equivalent to the FISTA algorithm. 


    """

    def __init__(self, initial, f, g, step_size=None, preconditioner=None, momentum=None,  **kwargs):

        super(APGD, self).__init__(**kwargs)

        self.set_up(initial=initial, f=f, g=g, step_size=step_size,
                    preconditioner=preconditioner, momentum=momentum)

    def set_up(self, initial, f, g, step_size, preconditioner, momentum):
        """Set up of the algorithm"""
        log.info("%s setting up", self.__class__.__name__)

        self.initial = initial
        self.y = initial.copy()
        self.x_old = initial.copy()
        self.x = initial.copy()
        self.gradient_update = initial.copy()

        if f is None:
            f = ZeroFunction()

        self.f = f

        if g is None:
            g = ZeroFunction()

        self.g = g

        if isinstance(f, ZeroFunction) and isinstance(g, ZeroFunction):
            raise ValueError(
                'You set both f and g to be the ZeroFunction and thus the iterative method will not update and will remain fixed at the initial value.')

        # set step_size
        if step_size is None:
            self.step_size_rule = ConstantStepSize(
                self._calculate_default_step_size())
        elif isinstance(step_size, Real):
            self.step_size_rule = ConstantStepSize(step_size)
        elif isinstance(step_size, StepSizeRule):
            self.step_size_rule = step_size
        else:
            raise TypeError(
                "Step_size must be a real number or a child class of :meth:`cil.optimisation.utilities.StepSizeRule`")

        self.preconditioner = preconditioner
        self._set_momentum(momentum)

        self.configured = True
        log.info("%s configured", self.__class__.__name__)

    def _set_momentum(self, momentum):

        if momentum is None:
            self._momentum = NesterovMomentum()
        else:
            if isinstance(momentum, Number):
                self._momentum = ConstantMomentum(momentum)
            elif isinstance(momentum, ScalarMomentumCoefficient):
                self._momentum = momentum
            else:
                raise TypeError(
                    "Momentum must be a number or a child class of ScalarMomentumCoefficient")

    def _calculate_default_step_size(self):
        """Calculate the default step size if a step size rule or step size is not provided 
        """
        return 1./self.f.L

    def update(self):
        r"""Performs a single iteration of APGD. For :math:`k\geq 1`:

        .. math::
        
                x_{k+1} = \mathrm{prox}_{\alpha g}(y_{k} - \alpha\nabla f(y_{k}))\\
            

        where :math:`\alpha` is the step size. From :math:`x_{k+1}` (and any other information available in the algorithm class) the momentum function then calculates :math:`y_{k+1}`. 
        """

        self.f.gradient(self.y, out=self.gradient_update)

        if self.preconditioner is not None:
            self.preconditioner.apply(
                self, self.gradient_update, out=self.gradient_update)

        self._step_size = self.step_size_rule.get_step_size(self)

        self.y.sapyb(1., self.gradient_update, -self._step_size, out=self.y)

        self.g.proximal(self.y, self._step_size, out=self.x)

        self.x.subtract(self.x_old, out=self.y)

        self.momentum.apply_momentum_in_APGD(self, out=self.y)

    def _update_previous_solution(self):
        """ Swaps the references to current and previous solution based on the :func:`~Algorithm.update_previous_solution` of the base class :class:`Algorithm`.
        """
        tmp = self.x_old
        self.x_old = self.x
        self.x = tmp

    def update_objective(self):
        """ Updates the objective

        .. math:: f(x) + g(x)

        """
        self.loss.append(
            self.calculate_objective_function_at_point(self.x_old))

    def calculate_objective_function_at_point(self, x):
        """ Calculates the objective at a given point x

        .. math:: f(x) + g(x)

        Parameters
        ----------
        x : DataContainer

        """
        return self.f(x) + self.g(x)

    def get_output(self):
        " Returns the current solution. "
        return self.x_old

    @property
    def step_size(self):
        '''
        Returns the most recently used step size. Note, if the step-size is set by a non-constant step size rule, you must use the algorithm run or update method before this getter will return the most recently used step size. 
        '''
        
        if isinstance(self.step_size_rule, ConstantStepSize):
            return self.step_size_rule.step_size
        else:
            try: 
                return self._step_size
            except AttributeError:
                raise NotImplementedError("Note the step-size is set by a step-size rule and could change with each iteration. Call the algorithm run or update method first and then this function will give the most recently used step size.")

    @property
    def momentum(self):
        return self._momentum

    def _provable_convergence_condition(self):
        if self.preconditioner is not None:
            raise NotImplementedError(
                "Can't check convergence criterion if a preconditioner is used ")

        if isinstance(self.step_size_rule, ConstantStepSize) and isinstance(self.momentum, NesterovMomentum):
            return self.step_size_rule.step_size <= 1./self.f.L
        else:
            raise TypeError(
                "Can't check convergence criterion for non-constant step size or non-Nesterov momentum coefficient")



#  Copyright 2018 United Kingdom Research and Innovation
#  Copyright 2018 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
import copy
from numbers import Number
import warnings

import numpy

from .labels import FillType


class BackwardCompat(type):
    @property
    def RANDOM(cls):
        warnings.warn("use FillType.RANDOM instead", DeprecationWarning, stacklevel=2)
        return FillType.RANDOM

    @property
    def RANDOM_INT(cls):
        warnings.warn("use FillType.RANDOM_INT instead", DeprecationWarning, stacklevel=2)
        return FillType.RANDOM_INT


class VectorGeometry:
    '''Geometry describing VectorData to contain 1D array'''
    @property
    def dtype(self):
        return self._dtype

    @dtype.setter
    def dtype(self, val):
        self._dtype = val

    def __init__(self,
                 length, **kwargs):

        self.length = int(length)
        self.shape = (length, )
        self.dtype = kwargs.get('dtype', numpy.float32)
        self.dimension_labels = kwargs.get('dimension_labels', None)

    def clone(self):
        '''returns a copy of VectorGeometry'''
        return copy.deepcopy(self)

    def copy(self):
        '''alias of clone'''
        return self.clone()

    def __eq__(self, other):

        if not isinstance(other, self.__class__):
            return False

        if self.length == other.length \
            and self.shape == other.shape \
            and self.dimension_labels == other.dimension_labels:
            return True
        return False

    def __str__ (self):
        repres = ""
        repres += "Length: {0}\n".format(self.length)
        repres += "Shape: {0}\n".format(self.shape)
        repres += "Dimension_labels: {0}\n".format(self.dimension_labels)

        return repres

    def allocate(self, value=0, **kwargs):
        '''Allocates a VectorData according to the geometry

        Parameters
        ----------
        value : number or string, default=0
            The value to allocate. Accepts a number to allocate a uniform array, 
            None to allocate an empty memory block, or a string to create a random 
            array: 'random' allocates floats between 0 and 1, 'random_int' allocates 
            ints between 0 and 100.

        **kwargs:
            dtype : numpy data type, optional
                The data type to allocate if different from the geometry data type. 
                Default None allocates an array with the geometry data type.

            seed : int, optional
                A random seed to fix reproducibility, only used if `value` is a random
                method. Default is `None`.

            min_value : number, optional
                The minimum value random integer to generate, only used if `value` 
                is 'random_int'. Default is 0.

            max_value : number, optional
                The maximum value random integer to generate, only used if `value` 
                is 'random_int'. Default is 100.

        Note
        ----
            The methods used by 'random' or 'random_int' use `numpy.random.default_rng` 
            which allocates memory only for the array of the specified dtype. This
            method does not use the global numpy.random.seed() so if a seed is 
            required it should be passed directly as an argument to allocate.
            To allocate random numbers using the deprecated `numpy.random.random_sample`
            and `numpy.random.randint` methods use `value='random_deprecated'` 
            or `value='random_int_deprecated'` 

        '''
        from .vector_data import VectorData

        dtype = kwargs.pop('dtype', self.dtype)

        out = VectorData(geometry=self.copy(), dtype=dtype)
        if value is not None:
            out.fill(value, **kwargs)

        return out



#  Copyright 2020 United Kingdom Research and Innovation
#  Copyright 2020 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.framework import ImageGeometry, AcquisitionGeometry, BlockGeometry
from cil.optimisation.operators import LinearOperator


class ChannelwiseOperator(LinearOperator):

    r'''The ChannelwiseOperator  takes in an operator  and the number of channels to be used, and creates a new multi-channel
    `ChannelwiseOperator`, which will apply the operator  independently on each channel for the number of channels specified.

    ChannelwiseOperator does not currently support BlockOperators. Typically, if such behaviour is desired, it can be achieved
    by creating instead a BlockOperator of ChannelwiseOperators.

    Parameters
    ----------
    op : Operator
        Single-channel operator
    channels : int
        Number of channels
    dimension : str, optional
        'prepend' (default) or 'append' channel dimension onto existing dimensions
        
    Example
    --------
    In this example, we create a ChannelwiseOperator that applies the same diagonal operator to each channel of a 2 channel image data.
    
    >>> M = 3
    >>> channels = 2
    >>> ig = ImageGeometry(M, M, channels=channels)
    >>> single_ig = ImageGeometry(M, M)
    >>> x = ImageData( np.stack( [np.ones((M,M)),  2*np.ones((M,M))] , axis=0), geometry=ig )
    >>> diag = ImageData(np.array(range(M*M), dtype=np.float64).reshape((M,M)), geometry=single_ig)
    >>> D = DiagonalOperator(diag)
    >>> C = ChannelwiseOperator(D,channels)
    >>> y = C.direct(x)
    >>> print('The original image data is:')
    >>> print(x.as_array())
    The original image data is:
    [[[1. 1. 1.]
    [1. 1. 1.]
    [1. 1. 1.]]
    [[2. 2. 2.]
    [2. 2. 2.]
    [2. 2. 2.]]]
    >>> print('The channel wise operator multiplies each channel element wise by:')
    >>> print(diag.as_array())
    The channel wise operator multiplies each channel element wise by:
    [[0. 1. 2.]
    [3. 4. 5.]
    [6. 7. 8.]]
    >>> print('The result of applying the channel wise operator is:')
    >>> print(y.as_array())
    The result of applying the channel wise operator is:
    [[[ 0.  1.  2.]
    [ 3.  4.  5.]
    [ 6.  7.  8.]]
    [[ 0.  2.  4.]
    [ 6.  8. 10.]
    [12. 14. 16.]]]
    
    '''

    def __init__(self, op, channels, dimension='prepend'):

        dom_op = op.domain_geometry()
        ran_op = op.range_geometry()

        geom_mc = []

        # Create multi-channel domain and range geometries: Clones of the
        # input single-channel geometries but with the specified number of
        # channels and additional dimension_label 'channel'.
        for geom in [dom_op,ran_op]:
            if dimension == 'prepend':
                new_dimension_labels = ['channel']+list(geom.dimension_labels)
            elif dimension == 'append':
                new_dimension_labels = list(geom.dimension_labels)+['channel']
            else:
                raise Exception("dimension must be either 'prepend' or 'append'")
            if isinstance(geom, ImageGeometry):

                geom_channels = geom.copy()
                geom_channels.channels = channels
                geom_channels.dimension_labels = new_dimension_labels

                geom_mc.append(geom_channels)
            elif isinstance(geom, AcquisitionGeometry):
                geom_channels = geom.copy()
                geom_channels.config.channels.num_channels = channels
                geom_channels.dimension_labels = new_dimension_labels

                geom_mc.append(geom_channels)

            elif isinstance(geom,BlockGeometry):
                raise Exception("ChannelwiseOperator does not support BlockOperator as input. Consider making a BlockOperator of ChannelwiseOperators instead.")
            else:
                pass

        super(ChannelwiseOperator, self).__init__(domain_geometry=geom_mc[0],
                                           range_geometry=geom_mc[1])

        self.op = op
        self.channels = channels

    def direct(self,x,out=None):
        '''
        Returns :math:`D(x)` where :math:`D` is the ChannelwiseOperator and :math:`x` is the input data.
        
        Parameters
        ----------
        x : DataContainer
            Input data
        out : DataContainer, optional
            Output data, if not provided a new DataContainer will be created
            
        Returns
        -------
        out : DataContainer
            Output data
        '''
        # Loop over channels, extract single-channel data, apply single-channel
        # operator's direct method and fill into multi-channel output data set.
        if out is None:
            out = self.range_geometry().allocate()
        cury = self.op.range_geometry().allocate()
        for k in range(self.channels):
            self.op.direct(x.get_slice(channel=k),cury)
            out.fill(cury.as_array(),channel=k)
        return out

    def adjoint(self, x, out=None):
        '''Returns :math:`D^{*}(x)` where :math:`D` is the ChannelwiseOperator and :math:`x` is the input data.
        
        Parameters
        ----------
        x : DataContainer
            Input data
        out : DataContainer, optional
            Output data, if not provided a new DataContainer will be created
    
        Returns
        -------
        out : DataContainer
            Output data
        
        '''
        # Loop over channels, extract single-channel data, apply single-channel
        # operator's adjoint method and fill into multi-channel output data set.
        if out is None:
            out = self.domain_geometry().allocate()
        cury = self.op.domain_geometry().allocate()
        for k in range(self.channels):
            self.op.adjoint(x.get_slice(channel=k),cury)
            out.fill(cury.as_array(),channel=k)
        return out

    def calculate_norm(self, **kwargs):
        '''Evaluates operator norm of ChannelWiseOperator'''
        return self.op.norm()



#  Copyright 2018 United Kingdom Research and Innovation
#  Copyright 2018 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
# Joshua DM Hellier (University of Manchester) [refactorer]
# Nicholas Whyatt (UKRI-STFC) [refactorer]
import copy
import ctypes
import warnings
from functools import reduce
from numbers import Number

import numpy

from .cilacc import cilacc
from cil.utilities.multiprocessing import NUM_THREADS


class DataContainer(object):
    '''Generic class to hold data

    Data is currently held in a numpy arrays'''

    @property
    def geometry(self):
        return None

    @geometry.setter
    def geometry(self, val):
        if val is not None:
            raise TypeError("DataContainers cannot hold a geometry, use ImageData or AcquisitionData instead")

    @property
    def dimension_labels(self):

        if self._dimension_labels is None:
            default_labels = [0]*self.number_of_dimensions
            for i in range(self.number_of_dimensions):
                default_labels[i] = 'dimension_{0:02}'.format(i)
            return tuple(default_labels)
        else:
            return self._dimension_labels

    @dimension_labels.setter
    def dimension_labels(self, val):
        if val is None:
            self._dimension_labels = None
        elif len(val_tuple := tuple(val)) == self.number_of_dimensions:
            self._dimension_labels = val_tuple
        else:
            raise ValueError("dimension_labels expected a list containing {0} strings got {1}".format(self.number_of_dimensions, val))

    @property
    def shape(self):
        '''Returns the shape of the DataContainer'''
        return self.array.shape

    @property
    def ndim(self):
        '''Returns the ndim of the DataContainer'''
        return self.array.ndim

    @property
    def number_of_dimensions(self):
        '''Returns the shape of the  of the DataContainer'''
        return len(self.array.shape)

    @property
    def dtype(self):
        '''Returns the dtype of the data array.'''
        return self.array.dtype

    @property
    def size(self):
        '''Returns the number of elements of the DataContainer'''
        return self.array.size

    __container_priority__ = 1
    def __init__ (self, array, deep_copy=True, dimension_labels=None,
                  **kwargs):
        if type(array) == numpy.ndarray:
            if deep_copy:
                self.array = array.copy()
            else:
                self.array = array
        else:
            raise TypeError('Array must be NumpyArray, passed {0}'\
                            .format(type(array)))

        #Don't set for derived classes
        if type(self) is DataContainer:
            self.dimension_labels = dimension_labels

        # finally copy the geometry, and force dtype of the geometry of the data = the dype of the data
        if 'geometry' in kwargs.keys():
            try:
                self.geometry = kwargs['geometry'].copy()
                if self.geometry.dtype != self.dtype:
                    warnings.warn("Over-riding geometry.dtype with data.dtype", UserWarning)
                    self.geometry.dtype = self.dtype
            except:
                pass

    def get_dimension_size(self, dimension_label):

        if dimension_label in self.dimension_labels:
            i = self.dimension_labels.index(dimension_label)
            return self.shape[i]
        else:
            raise ValueError('Unknown dimension {0}. Should be one of {1}'.format(dimension_label,
                             self.dimension_labels))

    def get_dimension_axis(self, dimension_label):
        """
        Returns the axis index of the DataContainer array if the specified dimension_label(s) match
        any dimension_labels of the DataContainer or their indices

        Parameters
        ----------
        dimension_label: string or int or tuple of strings or ints
            Specify dimension_label(s) or index of the DataContainer from which to check and return the axis index

        Returns
        -------
        int or tuple of ints
            The axis index of the DataContainer matching the specified dimension_label
        """
        if isinstance(dimension_label,(tuple,list)):
            return tuple(self.get_dimension_axis(x) for x in dimension_label)

        if dimension_label in self.dimension_labels:
            return self.dimension_labels.index(dimension_label)
        elif isinstance(dimension_label, int) and dimension_label >= 0 and dimension_label < self.ndim:
            return dimension_label
        else:
            raise ValueError('Unknown dimension {0}. Should be one of {1}, or an integer in range {2} - {3}'.format(dimension_label,
                            self.dimension_labels, 0, self.ndim))


    def as_array(self):
        '''Returns the pointer to the array.
        '''
        return self.array


    def get_slice(self, **kw):
        '''
        Returns a new DataContainer containing a single slice in the requested direction. \
        Pass keyword arguments <dimension label>=index
        '''
        # Force is not relevant for a DataContainer:
        kw.pop('force', None)

        new_array = None

        #get ordered list of current dimensions
        dimension_labels_list = list(self.dimension_labels)

        #remove axes from array and labels
        for key, value in kw.items():
            if value is not None:
                axis = dimension_labels_list.index(key)
                dimension_labels_list.remove(key)
                if new_array is None:
                    new_array = self.as_array()
                new_array = new_array.take(indices=value, axis=axis)

        if new_array.ndim > 1:
            return DataContainer(new_array, False, dimension_labels_list)
        from .vector_data import VectorData
        return VectorData(new_array, dimension_labels=dimension_labels_list)

    def reorder(self, order):
        '''
        Reorders the data in memory as requested. This is an in-place operation.

        Parameters
        ----------
        order: list or string
            ordered list of labels from self.dimension_labels
        '''
        try:
            if len(order) != len(self.shape):
                raise ValueError('The axes list for resorting must have {0} dimensions. Got {1}'.format(len(self.shape), len(order)))
        except TypeError as ae:
            raise ValueError('The order must be an iterable with __len__ implemented, like a list or a tuple. Got {}'.format(type(order)))

        correct = True
        for el in order:
            correct = correct and el in self.dimension_labels
        if not correct:
            raise ValueError('The axes list for resorting must contain the dimension_labels {0} got {1}'.format(self.dimension_labels, order))

        new_order = [0]*len(self.shape)
        dimension_labels_new = [0]*len(self.shape)

        for i, axis in enumerate(order):
            new_order[i] = self.dimension_labels.index(axis)
            dimension_labels_new[i] = axis

        self.array = numpy.ascontiguousarray(numpy.transpose(self.array, new_order))

        if self.geometry is None:
            self.dimension_labels = dimension_labels_new
        else:
            self.geometry.set_labels(dimension_labels_new)

    def fill(self, array, **kwargs):
        '''Fills the internal data array with a DataContainer, numpy array, number 
        or using a random fill method

        Parameters
        ----------
        array : DataContainer, numpy array, number or string
            The value or method with which to fill the DataContainer. Accepts a 
            numpy array or DataContainer or a number to allocate a uniform array
            or a string specifying a method to fill with a random array: 'random' 
            allocates floats between 0 and 1, 'random_int' by default allocates 
            integers between 0 and 100  or between provided `min_value` and `max_value`.

        **kwargs:
            **dimension : int, optional 
                An optional named parameter to specify in which axis the fill should 
                happen. The name passed must match one of the DataContainer 
                dimension_labels and the axis must be an int

            seed : int, optional
                A random seed to fix reproducibility, only used if `array` is a 
                `random method`. Default is `None`.

            min_value : int, optional, default=0
                The minimum value random integer to generate, only used if `array` 
                is 'random_int'. New since version 25.0.0 Default is 0.

            max_value : int, optional, default=100
                The maximum value random integer to generate, only used if `array` 
                is 'random_int'. Default is 100.
        
        Note
        ----
            If the passed numpy array points to the same array that is contained 
            in the DataContainer, the DataContainer is not updated, and None is returned.

        Note
        ----
            Since v25.0.0 the methods used by 'random' or 'random_int' use `numpy.random.default_rng`. 
            This method does not use the global numpy.random.seed() so if a seed is 
            required it should be passed directly as a kwarg.
            To fill random numbers using the earlier behaviour use `array='random_deprecated'` 
            or `array='random_int_deprecated'` 
        
        Example
        -------
        This example shows how to pass a dimension to specify an axis to fill
        >>> from cil.framework import DataContainer
        >>> DC = DataContainer(old_data)
        >>> DC.fill(new_data, vertical=1, horizontal_x=32)
        will copy the data in new_data into the data container.
        '''
        from cil.framework.labels import FillType
        dimension = {}

        try:
            dimension_labels = self.dimension_labels
        except AttributeError:
            dimension_labels = None
        
        if dimension_labels is not None:
            for k in list(kwargs):
                if k in self.dimension_labels:
                    dimension[k] = kwargs.pop(k)

            index = [slice(None)] * self.number_of_dimensions
            dimension_labels = tuple(self.dimension_labels)

            for k, v in dimension.items():
                i = dimension_labels.index(k)
                index[i] = v  
            index = tuple(index)
        else:
            index = (slice(None),) * self.array.ndim

        if id(array) == id(self.array):
            return
        
        if isinstance(array, numpy.ndarray):
            numpy.copyto(self.array[index], array)
        
        elif isinstance(array, Number):
            self.array[index] = array
        
        elif issubclass(array.__class__ , DataContainer):
            if dimension_labels is not None:
                indexed_dimension_labels = [label for label in self.dimension_labels if label not in dimension]
                if tuple(indexed_dimension_labels) != array.dimension_labels:
                    raise ValueError('Input array is not in the same order as destination array. Use "array.reorder()"')
            
            if self.array[index].shape == array.shape:
                numpy.copyto(self.array[index], array.array)
            else:
                raise ValueError('Cannot fill with the provided array.' + \
                                    'Expecting shape {0} got {1}'.format(
                                    self.array[index].shape, array.shape))
            
        elif array in FillType:

            seed = kwargs.pop("seed", None)
            
            if array == FillType.RANDOM_DEPRECATED:
                warnings.warn("RANDOM_DEPRECATED will be removed in a future version", DeprecationWarning, stacklevel=2)
                if seed is not None:
                    numpy.random.seed(seed)
                if numpy.issubdtype(self.dtype, numpy.complexfloating):
                    r = (numpy.random.random_sample(self.array[index].shape) + 1j * numpy.random.random_sample(self.array[index].shape)).astype(self.dtype)
                else:
                    r = numpy.random.random_sample(self.array[index].shape).astype(self.dtype)
                self.array[index] = r
            
            elif array == FillType.RANDOM_INT_DEPRECATED:
                warnings.warn("RANDOM_INT_DEPRECATED will be removed in a future version", DeprecationWarning, stacklevel=2)
                if seed is not None:
                    numpy.random.seed(seed)
                max_value = kwargs.pop("max_value", 100)
                min_value = kwargs.pop("min_value", 0)
                if numpy.issubdtype(self.dtype, numpy.complexfloating):
                    r = (numpy.random.randint(min_value, max_value,size=self.array[index].shape, dtype=numpy.int32) + 1j*numpy.random.randint(max_value,size=self.array[index].shape, dtype=numpy.int32)).astype(self.dtype)
                else:
                    r = numpy.random.randint(min_value, max_value,size=self.array[index].shape, dtype=numpy.int32).astype(self.dtype)
                self.array[index] = r

            elif array == FillType.RANDOM:
                
                rng = numpy.random.default_rng(seed)
                if numpy.issubdtype(self.dtype, numpy.complexfloating):
                    complex_example = numpy.array([1 + 1j], dtype=self.dtype)
                    half_dtype = numpy.real(complex_example).dtype
                    r = (rng.random(size=self.array[index].shape, dtype=half_dtype) + 1j * rng.random(size=self.array[index].shape, dtype=half_dtype)).astype(self.dtype)
                else:
                    r = rng.random(size=self.array[index].shape, dtype=self.dtype)
                self.array[index] = r

            elif array == FillType.RANDOM_INT:
                
                rng = numpy.random.default_rng(seed)
                max_value = kwargs.pop("max_value", 100)
                min_value = kwargs.pop("min_value", 0)
                if numpy.issubdtype(self.dtype, numpy.complexfloating):
                    r = (rng.integers(min_value, max_value, size=self.array[index].shape, dtype=numpy.int32) + 1j*rng.integers(0, max_value, size=self.array[index].shape, dtype=numpy.int32)).astype(self.dtype)
                else:
                    r = rng.integers(min_value, max_value, size=self.array[index].shape, dtype=numpy.int32).astype(self.dtype)
                self.array[index] = r

        else:
            raise TypeError('Can fill only with random method, number, numpy array or DataContainer and subclasses. Got {}'.format(type(array)))

        if kwargs:
            warnings.warn(f"Unused keyword arguments: {kwargs}", stacklevel=2)

    def check_dimensions(self, other):
        return self.shape == other.shape

    ## algebra

    def __add__(self, other):
        return self.add(other)
    def __mul__(self, other):
        return self.multiply(other)
    def __sub__(self, other):
        return self.subtract(other)
    def __div__(self, other):
        return self.divide(other)
    def __truediv__(self, other):
        return self.divide(other)
    def __pow__(self, other):
        return self.power(other)


    # reverse operand
    def __radd__(self, other):
        return self + other
    # __radd__

    def __rsub__(self, other):
        return (-1 * self) + other
    # __rsub__

    def __rmul__(self, other):
        return self * other
    # __rmul__

    def __rdiv__(self, other):
        tmp = self.power(-1)
        tmp *= other
        return tmp
    # __rdiv__
    def __rtruediv__(self, other):
        return self.__rdiv__(other)

    def __rpow__(self, other):
        if isinstance(other, Number) :
            fother = numpy.ones(numpy.shape(self.array)) * other
            return type(self)(fother ** self.array ,
                           dimension_labels=self.dimension_labels,
                           geometry=self.geometry)
    # __rpow__

    # in-place arithmetic operators:
    # (+=, -=, *=, /= , //=,
    # must return self

    def __iadd__(self, other):
        kw = {'out':self}
        return self.add(other, **kw)

    def __imul__(self, other):
        kw = {'out':self}
        return self.multiply(other, **kw)

    def __isub__(self, other):
        kw = {'out':self}
        return self.subtract(other, **kw)

    def __idiv__(self, other):
        kw = {'out':self}
        return self.divide(other, **kw)

    def __itruediv__(self, other):
        kw = {'out':self}
        return self.divide(other, **kw)

    def __neg__(self):
        '''negation operator'''
        return -1 * self

    def __str__ (self, representation=False):
        repres = ""
        repres += "Number of dimensions: {0}\n".format(self.number_of_dimensions)
        repres += "Shape: {0}\n".format(self.shape)
        repres += "Axis labels: {0}\n".format(self.dimension_labels)
        if representation:
            repres += "Representation: \n{0}\n".format(self.array)
        return repres

    def get_data_axes_order(self,new_order=None):
        '''returns the axes label of self as a list

        If new_order is None returns the labels of the axes as a sorted-by-key list.
        If new_order is a list of length number_of_dimensions, returns a list
        with the indices of the axes in new_order with respect to those in
        self.dimension_labels: i.e.
          >>> self.dimension_labels = {0:'horizontal',1:'vertical'}
          >>> new_order = ['vertical','horizontal']
          returns [1,0]
        '''
        if new_order is None:
            return self.dimension_labels
        else:
            if len(new_order) == self.number_of_dimensions:

                axes_order = [0]*len(self.shape)
                for i, axis in enumerate(new_order):
                    axes_order[i] = self.dimension_labels.index(axis)
                return axes_order
            else:
                raise ValueError(f"Expecting {len(self.shape)} axes, got {len(new_order)}")

    def clone(self):
        '''returns a copy of DataContainer'''
        return copy.deepcopy(self)

    def copy(self):
        '''alias of clone'''
        return self.clone()

    ## binary operations

    def pixel_wise_binary(self, pwop, x2, *args,  **kwargs):
        out = kwargs.get('out', None)

        if out is None:
            if isinstance(x2, Number):
                out = pwop(self.as_array() , x2 , *args, **kwargs )
            elif issubclass(x2.__class__ , DataContainer):
                out = pwop(self.as_array() , x2.as_array() , *args, **kwargs )
            elif isinstance(x2, numpy.ndarray):
                out = pwop(self.as_array() , x2 , *args, **kwargs )
            else:
                raise TypeError('Expected x2 type as number or DataContainer, got {}'.format(type(x2)))
            
            if self.geometry is None:
                return type(self)(out,
                    deep_copy=False,
                    dimension_labels=self.dimension_labels)
            
            return type(self)(out,
                deep_copy=False,
                geometry=self.geometry)


        elif issubclass(type(out), DataContainer) and issubclass(type(x2), DataContainer):
            if self.check_dimensions(out) and self.check_dimensions(x2):
                kwargs['out'] = out.as_array()
                pwop(self.as_array(), x2.as_array(), *args, **kwargs )
                #return type(self)(out.as_array(),
                #       deep_copy=False,
                #       dimension_labels=self.dimension_labels,
                #       geometry=self.geometry)
                return out
            raise ValueError(f"Wrong size for data memory: out {out.shape} x2 {x2.shape} expected {self.shape}")
        elif issubclass(type(out), DataContainer) and \
             isinstance(x2, (Number, numpy.ndarray)):
            if self.check_dimensions(out):
                if isinstance(x2, numpy.ndarray) and\
                    not (x2.shape == self.shape and x2.dtype == self.dtype):
                    raise ValueError(f"Wrong size for data memory: out {out.shape} x2 {x2.shape} expected {self.shape}")
                kwargs['out']=out.as_array()
                pwop(self.as_array(), x2, *args, **kwargs )
                return out
            raise ValueError(f"Wrong size for data memory: {out.shape} {self.shape}")
        elif issubclass(type(out), numpy.ndarray):
            if self.array.shape == out.shape and self.array.dtype == out.dtype:
                kwargs['out'] = out
                pwop(self.as_array(), x2, *args, **kwargs)
                #return type(self)(out,
                #       deep_copy=False,
                #       dimension_labels=self.dimension_labels,
                #       geometry=self.geometry)
        else:
            raise ValueError(f"incompatible class: {pwop.__name__} {type(out)}")

    def add(self, other, *args, **kwargs):
        if hasattr(other, '__container_priority__') and \
           self.__class__.__container_priority__ < other.__class__.__container_priority__:
            return other.add(self, *args, **kwargs)
        return self.pixel_wise_binary(numpy.add, other, *args, **kwargs)

    def subtract(self, other, *args, **kwargs):
        if hasattr(other, '__container_priority__') and \
           self.__class__.__container_priority__ < other.__class__.__container_priority__:
            return other.sapyb(-1,self,1, out=kwargs.get('out', None))
        return self.pixel_wise_binary(numpy.subtract, other, *args, **kwargs)

    def multiply(self, other, *args, **kwargs):
        if hasattr(other, '__container_priority__') and \
           self.__class__.__container_priority__ < other.__class__.__container_priority__:
            return other.multiply(self, *args, **kwargs)
        return self.pixel_wise_binary(numpy.multiply, other, *args, **kwargs)

    def divide(self, other, *args, **kwargs):
        if hasattr(other, '__container_priority__') and \
           self.__class__.__container_priority__ < other.__class__.__container_priority__:
            _out = other.divide(self, *args, **kwargs)
            _out.power(-1, out=_out)
            return _out
        return self.pixel_wise_binary(numpy.divide, other, *args, **kwargs)

    def power(self, other, *args, **kwargs):
        return self.pixel_wise_binary(numpy.power, other, *args, **kwargs)

    def maximum(self, x2, *args, **kwargs):
        return self.pixel_wise_binary(numpy.maximum, x2, *args, **kwargs)

    def minimum(self,x2, out=None, *args, **kwargs):
        return self.pixel_wise_binary(numpy.minimum, x2=x2, out=out, *args, **kwargs)


    def sapyb(self, a, y, b, out=None, num_threads=NUM_THREADS):
        '''performs a*self + b * y. Can be done in-place

        Parameters
        ----------
        a : multiplier for self, can be a number or a numpy array or a DataContainer
        y : DataContainer
        b : multiplier for y, can be a number or a numpy array or a DataContainer
        out : return DataContainer, if None a new DataContainer is returned, default None.
            out can be self or y.
        num_threads : number of threads to use during the calculation, using the CIL C library
            It will try to use the CIL C library and default to numpy operations, in case the C library does not handle the types.


        Example
        -------

        >>> a = 2
        >>> b = 3
        >>> ig = ImageGeometry(10,11)
        >>> x = ig.allocate(1)
        >>> y = ig.allocate(2)
        >>> out = x.sapyb(a,y,b)
        '''

        if out is None:
            out = self * 0.

        if out.dtype in [ numpy.float32, numpy.float64 ]:
            # handle with C-lib _axpby
            try:
                self._axpby(a, b, y, out, out.dtype, num_threads)
                return out
            except RuntimeError as rte:
                warnings.warn("sapyb defaulting to Python due to: {}".format(rte))
            except TypeError as te:
                warnings.warn("sapyb defaulting to Python due to: {}".format(te))
            finally:
                pass


        # cannot be handled by _axpby
        ax = self * a
        y.multiply(b, out=out)
        out.add(ax, out=out)
        return out

    def _axpby(self, a, b, y, out, dtype=numpy.float32, num_threads=NUM_THREADS):
        '''performs axpby with cilacc C library, can be done in-place.

        Does the operation .. math:: a*x+b*y and stores the result in out, where x is self

        :param a: scalar
        :type a: float
        :param b: scalar
        :type b: float
        :param y: DataContainer
        :param out: DataContainer instance to store the result
        :param dtype: data type of the DataContainers
        :type dtype: numpy type, optional, default numpy.float32
        :param num_threads: number of threads to run on
        :type num_threads: int, optional, default 1/2 CPU of the system
        '''

        c_float_p = ctypes.POINTER(ctypes.c_float)
        c_double_p = ctypes.POINTER(ctypes.c_double)

        #convert a and b to numpy arrays and get the reference to the data (length = 1 or ndx.size)
        try:
            nda = a.as_array()
        except:
            nda = numpy.asarray(a)

        try:
            ndb = b.as_array()
        except:
            ndb = numpy.asarray(b)

        a_vec = 0
        if nda.size > 1:
            a_vec = 1

        b_vec = 0
        if ndb.size > 1:
            b_vec = 1

        # get the reference to the data
        ndx = self.as_array()
        ndy = y.as_array()
        ndout = out.as_array()

        if ndout.dtype != dtype:
            raise Warning("out array of type {0} does not match requested dtype {1}. Using {0}".format(ndout.dtype, dtype))
            dtype = ndout.dtype
        if ndx.dtype != dtype:
            ndx = ndx.astype(dtype, casting='safe')
        if ndy.dtype != dtype:
            ndy = ndy.astype(dtype, casting='safe')
        if nda.dtype != dtype:
            nda = nda.astype(dtype, casting='same_kind')
        if ndb.dtype != dtype:
            ndb = ndb.astype(dtype, casting='same_kind')

        if dtype == numpy.float32:
            x_p = ndx.ctypes.data_as(c_float_p)
            y_p = ndy.ctypes.data_as(c_float_p)
            out_p = ndout.ctypes.data_as(c_float_p)
            a_p = nda.ctypes.data_as(c_float_p)
            b_p = ndb.ctypes.data_as(c_float_p)
            f = cilacc.saxpby

        elif dtype == numpy.float64:
            x_p = ndx.ctypes.data_as(c_double_p)
            y_p = ndy.ctypes.data_as(c_double_p)
            out_p = ndout.ctypes.data_as(c_double_p)
            a_p = nda.ctypes.data_as(c_double_p)
            b_p = ndb.ctypes.data_as(c_double_p)
            f = cilacc.daxpby

        else:
            raise TypeError('Unsupported type {}. Expecting numpy.float32 or numpy.float64'.format(dtype))

        #out = numpy.empty_like(a)


        # int psaxpby(float * x, float * y, float * out, float a, float b, long size)
        cilacc.saxpby.argtypes = [ctypes.POINTER(ctypes.c_float),  # pointer to the first array
                                  ctypes.POINTER(ctypes.c_float),  # pointer to the second array
                                  ctypes.POINTER(ctypes.c_float),  # pointer to the third array
                                  ctypes.POINTER(ctypes.c_float),  # pointer to A
                                  ctypes.c_int,                    # type of type of A selector (int)
                                  ctypes.POINTER(ctypes.c_float),  # pointer to B
                                  ctypes.c_int,                    # type of type of B selector (int)
                                  ctypes.c_longlong,               # type of size of first array
                                  ctypes.c_int]                    # number of threads
        cilacc.daxpby.argtypes = [ctypes.POINTER(ctypes.c_double), # pointer to the first array
                                  ctypes.POINTER(ctypes.c_double), # pointer to the second array
                                  ctypes.POINTER(ctypes.c_double), # pointer to the third array
                                  ctypes.POINTER(ctypes.c_double), # type of A (c_double)
                                  ctypes.c_int,                    # type of type of A selector (int)
                                  ctypes.POINTER(ctypes.c_double), # type of B (c_double)
                                  ctypes.c_int,                    # type of type of B selector (int)
                                  ctypes.c_longlong,               # type of size of first array
                                  ctypes.c_int]                    # number of threads

        if f(x_p, y_p, out_p, a_p, a_vec, b_p, b_vec, ndx.size, num_threads) != 0:
            raise RuntimeError('axpby execution failed')


    ## unary operations
    def pixel_wise_unary(self, pwop, *args,  **kwargs):
        out = kwargs.get('out', None)
        if out is None:
            out = pwop(self.as_array() , *args, **kwargs )

            if self.geometry is None:
                return type(self)(out,
                    deep_copy=False,
                    dimension_labels=self.dimension_labels)
            
            return type(self)(out,
                       deep_copy=False,
                       geometry=self.geometry)
        
        elif issubclass(type(out), DataContainer):
            if self.check_dimensions(out):
                kwargs['out'] = out.as_array()
                pwop(self.as_array(), *args, **kwargs )
            else:
                raise ValueError(f"Wrong size for data memory: {out.shape} {self.shape}")
        elif issubclass(type(out), numpy.ndarray):
            if self.array.shape == out.shape and self.array.dtype == out.dtype:
                kwargs['out'] = out
                pwop(self.as_array(), *args, **kwargs)
        else:
            raise ValueError("incompatible class: {pwop.__name__} {type(out)}")

    def abs(self, *args,  **kwargs):
        return self.pixel_wise_unary(numpy.abs, *args,  **kwargs)

    def sign(self, *args,  **kwargs):
        return self.pixel_wise_unary(numpy.sign, *args,  **kwargs)

    def sqrt(self, *args,  **kwargs):
        return self.pixel_wise_unary(numpy.sqrt, *args,  **kwargs)

    def conjugate(self, *args,  **kwargs):
        return self.pixel_wise_unary(numpy.conjugate, *args,  **kwargs)

    def exp(self, *args, **kwargs):
        '''Applies exp pixel-wise to the DataContainer'''
        return self.pixel_wise_unary(numpy.exp, *args, **kwargs)

    def log(self, *args, **kwargs):
        '''Applies log pixel-wise to the DataContainer'''
        return self.pixel_wise_unary(numpy.log, *args, **kwargs)

    ## reductions
    def squared_norm(self, **kwargs):
        '''return the squared euclidean norm of the DataContainer viewed as a vector'''
        #shape = self.shape
        #size = reduce(lambda x,y:x*y, shape, 1)
        #y = numpy.reshape(self.as_array(), (size, ))
        return self.dot(self)
        #return self.dot(self)
    def norm(self, **kwargs):
        '''return the euclidean norm of the DataContainer viewed as a vector'''
        return numpy.sqrt(self.squared_norm(**kwargs))

    def dot(self, other, *args, **kwargs):
        '''returns the inner product of 2 DataContainers viewed as vectors. Suitable for real and complex data.
          For complex data,  the dot method returns a.dot(b.conjugate())
        '''
        method = kwargs.get('method', 'numpy')
        if method not in ['numpy','reduce']:
            raise ValueError('dot: specified method not valid. Expecting numpy or reduce got {} '.format(
                    method))

        if self.shape == other.shape:
            if method == 'numpy':
                return numpy.dot(self.as_array().ravel(), other.as_array().ravel().conjugate())
            elif method == 'reduce':
                # see https://github.com/vais-ral/CCPi-Framework/pull/273
                # notice that Python seems to be smart enough to use
                # the appropriate type to hold the result of the reduction
                sf = reduce(lambda x,y: x + y[0]*y[1],
                            zip(self.as_array().ravel(),
                                other.as_array().ravel().conjugate()),
                            0)
                return sf
        else:
            raise ValueError('Shapes are not aligned: {} != {}'.format(self.shape, other.shape))

    def _directional_reduction_unary(self, reduction_function, axis=None, out=None, *args, **kwargs):
        """
        Returns the result of a unary function, considering the direction from an axis argument to the function

        Parameters
        ----------
        reduction_function : function
            The unary function to be evaluated
        axis : string or tuple of strings or int or tuple of ints, optional
            Specify the axis or axes to calculate 'reduction_function' along. Can be specified as
            string(s) of dimension_labels or int(s) of indices
            Default None calculates the function over the whole array
        out: ndarray or DataContainer, optional
            Provide an object in which to place the result. The object must have the correct dimensions and
            (for DataContainers) the correct dimension_labels, but the type will be cast if necessary. See
            `Output type determination <https://numpy.org/doc/stable/user/basics.ufuncs.html#ufuncs-output-type>`_ for more details.
            Default is None

        Returns
        -------
        scalar or ndarray
            The result of the unary function
        """
        if axis is not None:
            axis = self.get_dimension_axis(axis)

        if out is None:
            result = reduction_function(self.as_array(), axis=axis, *args, **kwargs)
            if isinstance(result, numpy.ndarray):
                new_dimensions = numpy.array(self.dimension_labels)
                new_dimensions = numpy.delete(new_dimensions, axis)
                return DataContainer(result, dimension_labels=new_dimensions)
            else:
                return result
        else:
            if hasattr(out,'array'):
                out_arr = out.array
            else:
                out_arr = out

            reduction_function(self.as_array(), out=out_arr, axis=axis,  *args, **kwargs)

    def sum(self, axis=None, out=None, *args, **kwargs):
        """
        Returns the sum of values in the DataContainer

        Parameters
        ----------
        axis : string or tuple of strings or int or tuple of ints, optional
            Specify the axis or axes to calculate 'sum' along. Can be specified as
            string(s) of dimension_labels or int(s) of indices
            Default None calculates the function over the whole array
        out : ndarray or DataContainer, optional
            Provide an object in which to place the result. The object must have the correct dimensions and
            (for DataContainers) the correct dimension_labels, but the type will be cast if necessary. See
            `Output type determination <https://numpy.org/doc/stable/user/basics.ufuncs.html#ufuncs-output-type>`_ for more details.
            Default is None

        Returns
        -------
        scalar or DataContainer
            The sum as a scalar or inside a DataContainer with reduced dimension_labels
            Default is to accumulate and return data as float64 or complex128
        """
        if kwargs.get('dtype') is not None:
            warnings.warn("dtype is ignored (auto-using float64 or complex128)", DeprecationWarning, stacklevel=2)

        if numpy.isrealobj(self.array):
            kwargs['dtype'] = numpy.float64
        else:
            kwargs['dtype'] = numpy.complex128

        return self._directional_reduction_unary(numpy.sum, axis=axis, out=out, *args, **kwargs)

    def min(self, axis=None, out=None, *args, **kwargs):
        """
        Returns the minimum pixel value in the DataContainer

        Parameters
        ----------
        axis : string or tuple of strings or int or tuple of ints, optional
            Specify the axis or axes to calculate 'min' along. Can be specified as
            string(s) of dimension_labels or int(s) of indices
            Default None calculates the function over the whole array
        out : ndarray or DataContainer, optional
            Provide an object in which to place the result. The object must have the correct dimensions and
            (for DataContainers) the correct dimension_labels, but the type will be cast if necessary.  See
            `Output type determination <https://numpy.org/doc/stable/user/basics.ufuncs.html#ufuncs-output-type>`_ for more details.
            Default is None

        Returns
        -------
        scalar or DataContainer
            The min as a scalar or inside a DataContainer with reduced dimension_labels
        """
        return self._directional_reduction_unary(numpy.min, axis=axis, out=out, *args, **kwargs)

    def max(self, axis=None, out=None, *args, **kwargs):
        """
        Returns the maximum pixel value in the DataContainer

        Parameters
        ----------
        axis : string or tuple of strings or int or tuple of ints, optional
            Specify the axis or axes to calculate 'max' along. Can be specified as
            string(s) of dimension_labels or int(s) of indices
            Default None calculates the function over the whole array
        out : ndarray or DataContainer, optional
            Provide an object in which to place the result. The object must have the correct dimensions and
            (for DataContainers) the correct dimension_labels, but the type will be cast if necessary. See
            `Output type determination <https://numpy.org/doc/stable/user/basics.ufuncs.html#ufuncs-output-type>`_ for more details.
            Default is None

        Returns
        -------
        scalar or DataContainer
            The max as a scalar or inside a DataContainer with reduced dimension_labels
        """
        return self._directional_reduction_unary(numpy.max, axis=axis, out=out, *args, **kwargs)

    def mean(self, axis=None, out=None, *args, **kwargs):
        """
        Returns the mean pixel value of the DataContainer

        Parameters
        ----------
        axis : string or tuple of strings or int or tuple of ints, optional
            Specify the axis or axes to calculate 'mean' along. Can be specified as
            string(s) of dimension_labels or int(s) of indices
            Default None calculates the function over the whole array
        out : ndarray or DataContainer, optional
            Provide an object in which to place the result. The object must have the correct dimensions and
            (for DataContainers) the correct dimension_labels, but the type will be cast if necessary. See
            `Output type determination <https://numpy.org/doc/stable/user/basics.ufuncs.html#ufuncs-output-type>`_ for more details.
            Default is None

        Returns
        -------
        scalar or DataContainer
            The mean as a scalar or inside a DataContainer with reduced dimension_labels
            Default is to accumulate and return data as float64 or complex128
        """

        if kwargs.get('dtype') is not None:
            warnings.warn("dtype is ignored (auto-using float64 or complex128)", DeprecationWarning, stacklevel=2)

        if numpy.isrealobj(self.array):
            kwargs['dtype'] = numpy.float64
        else:
            kwargs['dtype'] = numpy.complex128

        return self._directional_reduction_unary(numpy.mean, axis=axis, out=out, *args, **kwargs)

    # Logic operators between DataContainers and floats
    def __le__(self, other):
        '''Returns boolean array of DataContainer less or equal than DataContainer/float'''
        if isinstance(other, DataContainer):
            return self.as_array()<=other.as_array()
        return self.as_array()<=other

    def __lt__(self, other):
        '''Returns boolean array of DataContainer less than DataContainer/float'''
        if isinstance(other, DataContainer):
            return self.as_array()<other.as_array()
        return self.as_array()<other

    def __ge__(self, other):
        '''Returns boolean array of DataContainer greater or equal than DataContainer/float'''
        if isinstance(other, DataContainer):
            return self.as_array()>=other.as_array()
        return self.as_array()>=other

    def __gt__(self, other):
        '''Returns boolean array of DataContainer greater than DataContainer/float'''
        if isinstance(other, DataContainer):
            return self.as_array()>other.as_array()
        return self.as_array()>other

    def __eq__(self, other):
        '''Returns boolean array of DataContainer equal to DataContainer/float'''
        if isinstance(other, DataContainer):
            return self.as_array()==other.as_array()
        return self.as_array()==other

    def __ne__(self, other):
        '''Returns boolean array of DataContainer negative to DataContainer/float'''
        if isinstance(other, DataContainer):
            return self.as_array()!=other.as_array()
        return self.as_array()!=other



#  Copyright 2024 United Kingdom Research and Innovation
#  Copyright 2024 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at:
# https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt

from cil.framework import Processor, AcquisitionData
from cil.framework.labels import AcquisitionDimension, AcquisitionType

import numpy as np
from scipy.fft import fft2
from scipy.fft import ifft2
from scipy.fft import ifftshift
from scipy import constants
from tqdm import tqdm
import logging

log = logging.getLogger(__name__)

class PaganinProcessor(Processor):

    r"""
    Processor to retrieve quantitative information from phase contrast images
    using the Paganin phase retrieval algorithm described in [1]

    Parameters
    ----------
    delta: float (optional)
        Real part of the deviation of the material refractive index from 1,
        where refractive index :math:`n = (1 - \delta) + i \beta` energy-
        dependent refractive index information for x-ray wavelengths can be
        found at [2], default is 1

    beta: float (optional)
        Complex part of the material refractive index, where refractive index
        :math:`n = (1 - \delta) + i \beta` energy-dependent refractive index
        information for x-ray wavelengths can be found at [2], default is 1e-2

    energy: float (optional)
        Energy of the incident photon, default is 40000

    energy_units: string (optional)
        Energy units, default is 'eV'

    full_retrieval : bool, optional
        If True, perform the full phase retrieval and return the thickness. If
        False, return a filtered image, default is True

    filter_type: string (optional)
        The form of the Paganin filter to use, either 'paganin_method'
        (default) or 'generalised_paganin_method' as described in [3]

    pad: int (optional)
        Number of pixels to pad the image in Fourier space to reduce aliasing,
        default is 0

    return_units: string (optional)
        The distance units to return the sample thickness in, must be one of
        'm', 'cm', 'mm' or 'um'. Only applies if full_retrieval=True (default
        is'cm')

    Returns
    -------
    AcquisitionData
        AcquisitionData corrected for phase effects, retrieved sample thickness
        or (if :code:`full_retrieval=False`) filtered data

    Example
    -------
    >>> processor = PaganinProcessor(delta=5, beta=0.05, energy=18000)
    >>> processor.set_input(data)
    >>> thickness = processor.get_output()

    Example
    -------
    >>> processor = PaganinProcessor(delta=1,beta=10e2, full_retrieval=False)
    >>> processor.set_input(data)
    >>> filtered_image = processor.get_output()

    Example
    -------
    >>> processor = PaganinProcessor()
    >>> processor.set_input(data)
    >>> thickness = processor.get_output(override_filter={'alpha':10})
    >>> phase_retrieved_image = thickness*processor.mu

    Notes
    -----
    This processor will work most efficiently using the cil data order with
    `data.reorder('cil')`

    Notes
    -----
    This processor uses the phase retrieval algorithm described by Paganin et
    al. [1] to retrieve the sample thickness

    .. math:: T(x,y) = - \frac{1}{\mu}\ln\left (\mathcal{F}^{-1}\left
        (\frac{\mathcal{F}\left ( M^2I_{norm}(x, y,z = \Delta) \right )}{1 +
          \alpha\left ( k_x^2 + k_y^2 \right )}  \right )\right ),

    where

        - :math:`T`, is the sample thickness,
        - :math:`\mu = \frac{4\pi\beta}{\lambda}` is the material linear
        attenuation coefficient where :math:`\beta` is the complex part of the
        material refractive index and :math:`\lambda=\frac{hc}{E}` is the probe
        wavelength,
        - :math:`M` is the magnification at the detector,
        - :math:`I_{norm}` is the input image which is expected to be the
        normalised transmission data,
        - :math:`\Delta` is the propagation distance,
        - :math:`\alpha = \frac{\Delta\delta}{\mu}` is a parameter determining
        the strength of the filter to be applied in Fourier space where
        :math:`\delta` is the real part of the deviation of the material
        refractive index from 1
        - :math:`k_x, k_y = \left ( \frac{2\pi p}{N_xW}, \frac{2\pi q}{N_yW}
        \right )` where :math:`p` and :math:`q` are co-ordinates in a Fourier
        mesh in the range :math:`-N_x/2` to :math:`N_x/2` for an image with
        size :math:`N_x, N_y` and pixel size :math:`W`.

    A generalised form of the Paganin phase retrieval method can be called
    using :code:`filter_type='generalised_paganin_method'`, which uses the
    form of the algorithm described in [2]

    .. math:: T(x,y) = -\frac{1}{\mu}\ln\left (\mathcal{F}^{-1}\left (\frac{
        \mathcal{F}\left ( M^2I_{norm}(x, y,z = \Delta) \right )}{1 - \frac{2
        \alpha}{W^2}\left ( \cos(Wk_x) + \cos(Wk_y) -2 \right )}  \right )
        \right )

    The phase retrieval is valid under the following assumptions

        - used with paraxial propagation-induced phase contrast images which
        can be assumed to be single-material locally
        - using intensity data which has been flat field corrected
        - and under the assumption that the Fresnel number
        :math:`F_N = W^2/(\lambda\Delta) >> 1`

    To apply a filter to images using the Paganin method, call
    :code:`full_retrieval=False`. In this case the pre-scaling and conversion
    to absorption is not applied so the requirement to supply flat field
    corrected intensity data is relaxed,

    .. math:: I_{filt} = \mathcal{F}^{-1}\left (\frac{\mathcal{F}\left (
        I(x, y,z = \Delta) \right )}
        {1 - \alpha\left ( k_x^2 + k_y^2 \right )}  \right )

    References
    ---------
    - [1] https://doi.org/10.1046/j.1365-2818.2002.01010.x
    - [2] https://henke.lbl.gov/optical_constants/getdb2.html
    - [3] https://iopscience.iop.org/article/10.1088/2040-8986/abbab9
    With thanks to colleagues at DTU for help with the initial implementation
    of the phase retrieval algorithm

    """

    def __init__(self, delta=1, beta=1e-2, energy=40000,
                 energy_units='eV',  full_retrieval=True,
                 filter_type='paganin_method', pad=0,
                 return_units='cm'):

        kwargs = {
            'energy' : energy,
            'wavelength' : self._energy_to_wavelength(energy, energy_units,
                                                      return_units),
            'delta': delta,
            'beta': beta,
            '_delta_user' : delta,
            '_beta_user' : beta,
            'filter_Nx' : None,
            'filter_Ny' : None,
            'filter_type' : filter_type,
            'mu' : None,
            'alpha' : None,
            'pixel_size' : None,
            'propagation_distance' : None,
            'magnification' : None,
            'filter' : None,
            'full_retrieval' : full_retrieval,
            'pad' : pad,
            'override_geometry' : None,
            'override_filter' : None,
            'return_units' : return_units
            }

        super(PaganinProcessor, self).__init__(**kwargs)

    def check_input(self, data):
        if not isinstance(data, (AcquisitionData)):
            raise TypeError('Processor only supports AcquisitionData')
        
        if data.geometry.geom_type & AcquisitionType.CONE_FLEX:
            raise NotImplementedError("Processor not implemented for CONE_FLEX geometry.")
        
        if data.dtype!=np.float32:
            raise TypeError('Processor only support dtype=float32')
        
        cil_order = tuple(AcquisitionDimension.get_order_for_engine('cil',data.geometry))
        if data.dimension_labels != cil_order:
            raise ValueError("This processor requires CIL data order, consider using `data.reorder('cil')`")
    
        return True

    def process(self, out=None):

        data  = self.get_input()            
        # set the geometry parameters to use from data.geometry unless the
        # geometry is overridden with an override_geometry
        self._set_geometry(data.geometry, self.override_geometry)

        if out is None:
            out = data.geometry.allocate(None)

        target_shape = (
            data.get_dimension_size('channel')  if 'channel' in data.dimension_labels else 1,
            data.geometry.num_projections,
            data.get_dimension_size('vertical') if 'vertical' in data.dimension_labels else 1, 
            data.get_dimension_size('horizontal') if 'horizontal' in data.dimension_labels else 1)
            
        data.array = np.reshape(data.array, target_shape)
        out.array = np.reshape(out.array, target_shape)
        
        # create a filter based on the shape of the data
        proj_shape = data.array.shape[-2:]
        self.filter_Nx = proj_shape[0]+self.pad*2
        self.filter_Ny = proj_shape[1]+self.pad*2
        self._create_filter(self.override_filter)

        # allocate padded buffer
        padded_buffer = np.zeros((self.filter_Nx, self.filter_Ny), dtype=data.dtype)
        buffer_slice = (slice(self.pad, self.pad + proj_shape[0]), 
                    slice(self.pad, self.pad + proj_shape[1]))
        
        # pre-calculate the scaling factor
        scaling_factor = -(1/self.mu)
        mag2 = self.magnification ** 2
        
        # loop over the channels
        for j in range(data.geometry.channels):
            # loop over the projections
            for i in tqdm(range(target_shape[1])):
                padded_buffer[:] = 0
                padded_buffer[buffer_slice] = data.array[j, i, :, :]

                if self.full_retrieval==True:
                    # apply the filter in fourier space, apply log and scale
                    padded_buffer*=mag2
                    fI = fft2(padded_buffer)
                    iffI = ifft2(fI*self.filter).real
                    np.log(iffI, out=padded_buffer)
                    np.multiply(scaling_factor, padded_buffer, out=padded_buffer)
                else:
                    # apply the filter in fourier space
                    fI = fft2(padded_buffer)
                    padded_buffer[:] = ifft2(fI*self.filter).real
                
                out.array[j, i, :, :] = padded_buffer[buffer_slice]
                        
        data.array = np.squeeze(data.array)
        out.array = np.squeeze(out.array)
        return out

    def set_input(self, dataset):
        """
        Set the input data to the processor

        Parameters
        ----------
        dataset : AcquisitionData
            The input AcquisitionData
        """
        return super().set_input(dataset)

    def get_output(self, out=None, override_geometry=None,
                   override_filter=None):
        r'''
        Function to get output from the PaganinProcessor

        Parameters
        ----------
        out : DataContainer, optional
           Fills the referenced DataContainer with the processed data

        override_geometry: dict, optional
            Geometry parameters to use in the phase retrieval if you want to
            over-ride values found in `data.geometry`. Specify parameters as a
            dictionary :code:`{'parameter':value}` where parameter is
            :code:`'magnification', 'propagation_distance'` or
            :code:`'pixel_size'` and value is the new value to use. Specify
            distance parameters in the same units as :code:`return_units`
            (default is cm).

        override_filter: dict, optional
            Over-ride the filter parameters to use in the phase retrieval.
            Specify parameters as :code:`{'parameter':value}` where parameter
            is :code:`'delta', 'beta'` or :code:`'alpha'` and value is the new
            value to use.

        Returns
        -------
        AcquisitionData
            AcquisitionData corrected for phase effects, retrieved sample
            thickness or (if :code:`full_retrieval=False`) filtered data

        Example
        -------
        >>> processor = PaganinProcessor(delta=5, beta=0.05, energy=18000)
        >>> processor.set_input(data)
        >>> thickness = processor.get_output()

        Example
        -------
        >>> processor = PaganinProcessor(delta=1,beta=10e2,
        full_retrieval=False)
        >>> processor.set_input(data)
        >>> filtered_image = processor.get_output()

        Example
        -------
        >>> processor = PaganinProcessor()
        >>> processor.set_input(data)
        >>> thickness = processor.get_output(override_filter={'alpha':10})
        >>> phase_retrieved_image = thickness*processor.mu

        Notes
        -----
        If :code:`'alpha'` is specified in override_filter the new value will
        be used and delta will be ignored but beta will still be used to
        calculate :math:`\mu = \frac{4\pi\beta}{\lambda}` which is used for
        scaling the thickness, therefore it is only recommended to specify
        alpha when also using :code:`get_output(full_retrieval=False)`, or
        re-scaling the result by :math:`\mu` e.g.
        :code:`thickness*processor.mu` If :code:`alpha` is not specified,
        it will be calculated :math:`\frac{\Delta\delta\lambda}{4\pi\beta}`

        '''
        self.override_geometry = override_geometry
        self.override_filter = override_filter

        return super().get_output(out)

    def __call__(self, x, out=None, override_geometry=None,
                 override_filter=None):
        self.set_input(x)

        if out is None:
            out = self.get_output(override_geometry=override_geometry,
                                  override_filter=override_filter)
        else:
            self.get_output(out=out, override_geometry=override_geometry,
                            override_filter=override_filter)

        return out

    def _set_geometry(self, geometry, override_geometry=None):
        '''
        Function to set the geometry parameters for the processor. Values are
        from the data geometry unless the geometry is overridden with an
        override_geometry dictionary.
        '''

        parameters = ['magnification', 'propagation_distance', 'pixel_size']
        # specify parameter names as defined in geometry
        geometry_parameters = ['magnification', 'dist_center_detector',
                               ('pixel_size_h', 'pixel_size_v')]
        # specify if parameter requires unit conversion
        convert_units = [False, True, True]

        if override_geometry is None:
            override_geometry = {}

        # get and check parameters from over-ride geometry dictionary
        for parameter in override_geometry.keys():
            if parameter not in parameters:
                raise ValueError('Parameter {} not recognised, expected one of\
                                 {}.'.format(parameter, parameters))
            elif (override_geometry[parameter] is None) \
                | (override_geometry[parameter] == 0):
                raise ValueError("Parameter {} cannot be {}, please update \
                                 data.geometry.{} or over-ride with \
                                 processor.get_output(override_geometry= \
                                 {{ '{}' : value }} )"\
                    .format(parameter, str(getattr(self, parameter)),
                            geometry_parameters[i], parameter))
            else:
                self.__setattr__(parameter, override_geometry[parameter])


        # get and check parameters from geometry if they are not in the
        # over-ride geometry dictionary
        for i, parameter in enumerate(parameters):
            if parameter not in override_geometry:
                if type(geometry_parameters[i])==tuple:
                    param1 = getattr(geometry, geometry_parameters[i][0])
                    param2 = getattr(geometry, geometry_parameters[i][1])
                    if (param1 - param2) / (param1 + param2) >= 1e-5:
                        raise ValueError("Parameter {} is not homogeneous up \
                                         to 1e-5: got {} and {}, please update\
                                          geometry using data.geometry.{} and \
                                         data.geometry.{} or over-ride with \
                                         processor.get_output(\
                                         override_geometry={{ '{}' : value }})"
                                         .format(parameter, str(param1),
                                                 str(param2),
                                                 geometry_parameters[i][0],
                                                 geometry_parameters[i][1],
                                                 parameter))
                else:
                    param1 = getattr(geometry, geometry_parameters[i])

                if (param1 is None) | (param1 == 0):
                    raise ValueError("Parameter {} cannot be {}, please update\
                                      data.geometry.{} or over-ride with \
                                     processor.get_output(override_geometry\
                                     ={{ '{}' : value }} )"
                                     .format(parameter, str(param1),
                                             str(geometry_parameters[i]),
                                             parameter))
                else:
                    if convert_units[i]:
                        param1 = self._convert_units(param1, 'distance',
                                                       geometry.config.units,
                                                       self.return_units)
                    self.__setattr__(parameter, param1)


    def _create_filter(self, override_filter=None):
        '''
        Function to create the Paganin filter, either using the paganin [1] or
        generalised paganin [2] method
        The filter is created on a mesh in Fourier space kx, ky
        [1] https://doi.org/10.1046/j.1365-2818.2002.01010.x
        [2] https://iopscience.iop.org/article/10.1088/2040-8986/abbab9
        '''
        if override_filter is None:
            override_filter = {}

        # update any parameter which has been over-ridden with override_filter
        if ('alpha' in override_filter) & ('delta' in override_filter):
            log.warning(msg="Because you specified alpha, it will not be \
                        calculated and therefore delta will be ignored")

        if ('delta' in override_filter):
            self.delta = override_filter['delta']
        else:
            self.delta = self._delta_user

        if ('beta' in override_filter):
            self.beta = override_filter['beta']
        else:
            self.beta = self._beta_user

        self._calculate_mu()

        if ('alpha' in override_filter):
            self.alpha = override_filter['alpha']
        else:
            self._calculate_alpha()

        # create the Fourier mesh
        kx,ky = np.meshgrid(
            np.arange(-self.filter_Nx/2, self.filter_Nx/2, 1, dtype=np.float64)
            * (2*np.pi)/(self.filter_Nx*self.pixel_size),
            np.arange(-self.filter_Ny/2, self.filter_Ny/2, 1, dtype=np.float64)
            * (2*np.pi)/(self.filter_Ny*self.pixel_size),
            sparse=False,
            indexing='ij'
            )

        # create the filter using either paganin or generalised paganin method
        if self.filter_type == 'paganin_method':
            self.filter =  ifftshift(1/(1. + self.alpha*(kx**2 + ky**2)))
        elif self.filter_type == 'generalised_paganin_method':
            self.filter =  ifftshift(1/(1. - (2*self.alpha/self.pixel_size**2)
                                        *(np.cos(self.pixel_size*kx)
                                          + np.cos(self.pixel_size*ky) -2)))
        else:
            raise ValueError("filter_type not recognised: got {0} expected one\
                              of 'paganin_method' or \
                             'generalised_paganin_method'"
                             .format(self.filter_type))

    def _calculate_mu(self):
        '''
        Function to calculate the linear attenutation coefficient mu
        '''
        self.mu = 4.0*np.pi*self.beta/self.wavelength

    def _calculate_alpha(self):
        '''
        Function to calculate alpha, a constant defining the Paganin filter
        strength
        '''
        self.alpha = self.propagation_distance*self.delta/self.mu

    def _energy_to_wavelength(self, energy, energy_units, return_units):
        '''
        Function to convert photon energy in eV to wavelength in return_units

        Parameters
        ----------
        energy: float
            Photon energy

        energy_units
            Energy units

        return_units
            Distance units in which to return the wavelength

        Returns
        -------
        float
            Photon wavelength in return_units
        '''
        top = self._convert_units(constants.h*constants.speed_of_light,
                                    'distance', 'm', return_units)
        bottom = self._convert_units(energy, 'energy', energy_units, 'J')

        return top/bottom

    def _convert_units(self, value, unit_type, input_unit, output_unit):
        unit_types = ['distance','energy','angle']

        if unit_type == unit_types[0]:
            unit_list = ['m','cm','mm','um']
            unit_multipliers = [1.0, 1e-2, 1e-3, 1e-6]
        elif unit_type == unit_types[1]:
            unit_list = ['meV', 'eV', 'keV', 'MeV', 'J']
            unit_multipliers = [1e-3, 1, 1e3, 1e6, 1/constants.eV]
        elif unit_type == unit_types[2]:
            unit_list = ['deg', 'rad']
            unit_multipliers = [1, np.rad2deg(1)]
        else:
            raise ValueError("Unit type '{}' not recognised, must be one of {}"
                            .format(unit_type, unit_types))

        for x in [input_unit, output_unit]:
            if x not in unit_list:
                raise ValueError("Unit '{}' not recognised, must be one of {}.\
                                 \nGeometry units can be updated using geometry.config.units"
                                 .format(x, unit_list))

        return value*unit_multipliers[unit_list.index(input_unit)]\
            /unit_multipliers[unit_list.index(output_unit)]



#  Copyright 2024 United Kingdom Research and Innovation
#  Copyright 2024 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# - CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt


from abc import ABC, abstractmethod
from functools import partialmethod

from tqdm.auto import tqdm as tqdm_auto
from tqdm.std import tqdm as tqdm_std
import numpy as np


class Callback(ABC):
    '''Base Callback to inherit from for use in :code:`Algorithm.run(callbacks: list[Callback])`.

    Parameters
    ----------
    verbose: int, choice of 0,1,2, default 1
        0=quiet, 1=info, 2=debug.
    '''
    def __init__(self, verbose=1):
        self.verbose = verbose

    @abstractmethod
    def __call__(self, algorithm):
        pass


class _OldCallback(Callback):
    '''Converts an old-style :code:`def callback` to a new-style :code:`class Callback`.

    Parameters
    ----------
    callback: :code:`callable(iteration, objective, x)`
    '''
    def __init__(self, callback, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.func = callback

    def __call__(self, algorithm):
        if algorithm.update_objective_interval > 0 and algorithm.iteration % algorithm.update_objective_interval == 0:
            self.func(algorithm.iteration, algorithm.get_last_objective(return_all=self.verbose>=2), algorithm.x)


class ProgressCallback(Callback):
    ''':code:`tqdm`-based progress bar.

    Parameters
    ----------
    tqdm_class: default :code:`tqdm.auto.tqdm`
    **tqdm_kwargs:
        Passed to :code:`tqdm_class`.
    '''
    def __init__(self, verbose=1, tqdm_class=tqdm_auto, **tqdm_kwargs):
        super().__init__(verbose=verbose)
        self.tqdm_class = tqdm_class
        self.tqdm_kwargs = tqdm_kwargs
        self._obj_len = 0  # number of objective updates

    def __call__(self, algorithm):
        if not hasattr(self, 'pbar'):
            tqdm_kwargs = self.tqdm_kwargs
            tqdm_kwargs.setdefault('total', algorithm.max_iteration)
            tqdm_kwargs.setdefault('disable', not self.verbose)
            tqdm_kwargs.setdefault('initial', max(0, algorithm.iteration))
            self.pbar = self.tqdm_class(**tqdm_kwargs)
        if (obj_len := len(algorithm.objective)) != self._obj_len:
            self.pbar.set_postfix(algorithm.objective_to_dict(self.verbose>=2), refresh=False)
            self._obj_len = obj_len
        self.pbar.update(algorithm.iteration - self.pbar.n)


class _TqdmText(tqdm_std):
    ''':code:`tqdm`-based progress but text-only updates on separate lines.

    Parameters
    ----------
    num_format: str
        Format spec for postfix numbers (i.e. objective values).
    bar_format: str
        Passed to :code:`tqdm`.
    '''
    def __init__(self, *args, num_format='+8.3e', bar_format="{n:>6d}/{total_fmt:<6} {rate_fmt:>9}{postfix}", **kwargs):
        self.num_format = num_format
        super().__init__(*args, bar_format=bar_format, mininterval=0, maxinterval=0, position=0, **kwargs)
        self._instances.remove(self)  # don't interfere with external progress bars

    @staticmethod
    def status_printer(file):
        fp_flush = getattr(file, 'flush', lambda: None)

        def fp_write(s):
            file.write(f"{s}\n")
            fp_flush()

        return fp_write

    def format_num(self, n):
        return f'{n:{self.num_format}}'

    def display(self, *args, **kwargs):
        """
        Clears :code:`postfix` if :code:`super().display()` succeeds
        (if display updates are more frequent than objective updates, users should not think the objective has stabilised).
        """
        if (updated := super().display(*args, **kwargs)):
            self.set_postfix_str('', refresh=False)
        return updated


class TextProgressCallback(ProgressCallback):
    ''':code:`ProgressCallback` but printed on separate lines to screen.

    Parameters
    ----------
    miniters: int, default :code:`Algorithm.update_objective_interval`
        Number of algorithm iterations between screen prints.
    '''
    __init__ = partialmethod(ProgressCallback.__init__, tqdm_class=_TqdmText)

    def __call__(self, algorithm):
        if not hasattr(self, 'pbar'):
            self.tqdm_kwargs['miniters'] = min((
                self.tqdm_kwargs.get('miniters', algorithm.update_objective_interval),
                algorithm.update_objective_interval))
        return super().__call__(algorithm)


class LogfileCallback(TextProgressCallback):
    ''':code:`TextProgressCallback` but to a file instead of screen.

    Parameters
    ----------
    log_file: FileDescriptorOrPath
        Passed to :code:`open()`.
    mode: str
        Passed to :code:`open()`.
    '''
    def __init__(self, log_file, mode='a', **kwargs):
        self.fd = open(log_file, mode=mode)
        super().__init__(file=self.fd, **kwargs)
        
class EarlyStoppingObjectiveValue(Callback):
    '''Callback that stops iterations if the change in the objective value is less than a provided threshold value.

    Parameters
    ----------
    threshold: float, default 1e-6 

    Note
    -----
    This callback only compares the last two calculated objective values. If `update_objective_interval` is greater than 1, the objective value is not calculated at each iteration (which is the default behaviour), only every `update_objective_interval` iterations.
    
    '''
    def __init__(self, threshold=1e-6):
        self.threshold=threshold
    
    
    def __call__(self, algorithm):
        if len(algorithm.loss)>=2:
            if np.abs(algorithm.loss[-1]-algorithm.loss[-2])<self.threshold:
                raise StopIteration
                
class CGLSEarlyStopping(Callback):
    '''Callback to work with CGLS. It causes the algorithm to terminate if  :math:`||A^T(Ax-b)||_2 < \epsilon||A^T(Ax_0-b)||_2` where `epsilon` is set to default as '1e-6', :math:`x` is the current iterate and :math:`x_0` is the initial value. 
    It will also terminate if the algorithm begins to diverge i.e. if :math:`||x||_2> \omega`, where `omega` is set to default as 1e6. 
    
    Parameters
    ----------
    epsilon: float, default 1e-6 
        Usually a small number: the algorithm to terminate if :math:`||A^T(Ax-b)||_2 < \epsilon||A^T(Ax_0-b)||_2`
    omega: float, default 1e6 
        Usually a large number: the algorithm will terminate if  :math:`||x||_2> \omega`
        
    Note
    -----
    This callback is implemented to replicate the automatic behaviour of CGLS in CIL versions <=24. It also replicates the behaviour of https://web.stanford.edu/group/SOL/software/cgls/. 
    '''
    def __init__(self, epsilon=1e-6, omega=1e6):
        self.epsilon=epsilon
        self.omega=omega
    
    
    def __call__(self, algorithm):
        
        if (algorithm.norms <= algorithm.norms0 * self.epsilon):
            print('The norm of the residual is less than {} times the norm of the initial residual and so the algorithm is terminated'.format(self.epsilon))
            raise StopIteration
        self.normx = algorithm.x.norm()
        if algorithm.normx >= self.omega:
            print('The norm of the solution is greater than {} and so the algorithm is terminated'.format(self.omega))
            raise StopIteration
            
        
